# -*- coding: utf-8 -*-
"""SuperDARN data support for grdex files(Alpha Level!)

Properties
----------
platform
    'superdarn'
name
    'grdex'
tag
    'north' or 'south' for Northern/Southern hemisphere data

Note
----
Requires davitpy and davitpy to load SuperDARN files.
Uses environment variables set by davitpy to download files
from Virginia Tech SuperDARN data servers. davitpy routines
are used to load SuperDARN data.

This material is based upon work supported by the
National Science Foundation under Grant Number 1259508.

Any opinions, findings, and conclusions or recommendations expressed in this
material are those of the author(s) and do not necessarily reflect the views
of the National Science Foundation.


Warnings
--------
Cleaning only removes entries that have 0 vectors, grdex files
are constituted from what it is thought to be good data.

"""

from __future__ import print_function
from __future__ import absolute_import
import functools

import pandas as pds
import numpy as np

import pysat

import logging
logger = logging.getLogger(__name__)


platform = 'superdarn'
name = 'grdex'
tags = {'north': '',
        'south': ''}
sat_ids = {'': ['north', 'south']}
_test_dates = {'': {'north': pysat.datetime(2009, 1, 1),
                    'south': pysat.datetime(2009, 1, 1)}}


def init(self):
    """Initializes the Instrument object with instrument specific values.

    Runs once upon instantiation.

    Parameters
    ----------
    self : pysat.Instrument
        This object

    Returns
    --------
    Void : (NoneType)
        Object modified in place.


    """

    # reset the list_remote_files routine to include the data path
    # now conveniently included with instrument object
    self._list_remote_rtn = \
        functools.partial(list_remote_files,
                          data_path=self.files.data_path,
                          format_str=self.files.file_format)

    # data acknowledgement from SuperDARN
    # coped from SD Documents area of VT SuperDARN webpage
    # http://vt.superdarn.org/tiki-list_file_gallery.php?galleryId=81
    # How to acknowledge use of SuperDARN Data - 2017
    logger.info('Authors should acknowledge the use of SuperDARN data. ' +
          'SuperDARN is a collection of radars funded by national scientific ' +
          'funding agencies of Australia, Canada, China, France, Italy, ' +
          'Japan, Norway, South Africa, United Kingdom and the United States ' +
          'of America.')
    return


def list_remote_files(tag, sat_id, data_path=None, format_str=None):
    """Lists remote files available for SuperDARN.

    Note
    ----
    This routine currently fakes the list but
    produces the desired effect of keeping data current.
    Begins with data in 1985. (this needs to be checked)

    Parameters
    ----------
    tag : (string or NoneType)
        Denotes type of file to load.  Accepted types are <tag strings>.
        (default=None)
    sat_id : (string or NoneType)
        Specifies the satellite ID for a constellation.  Not used.
        (default=None)

    Returns
    -------
    pandas.Series
        Series indexed by date that stores the filename for each date.

    """

    # given the function of SuperMAG, create a fake list of files
    # starting 01 Jan 1970, through today
    now = pysat.datetime.now()
    now = pysat.datetime(now.year, now.month, now.day)
    # create a list of dates with appropriate frequency
    index = pds.period_range(pysat.datetime(1985, 1, 1), now, freq='D')
    # pre fill in blank strings
    remote_files = pds.Series([''] * len(index), index=index)

    # pysat compares both dates and filenames when determining
    # which files it needs to download
    # so we need to ensure that filename for dates that overlap
    # are the same or data that is already present will be redownloaded

    # need to get a list of the current files attached to
    # the Instrument object. In this case, the object hasn't
    # been passed in.....
    #   that is ok, we can just call list_files right here
    #   except we don't have the data path
    # the init function above is used to reset the
    # lost_remote_files method with one where the
    # data path and format_str are set
    local_files = list_files(tag, sat_id, data_path, format_str)
    # iterating directly since pandas is complaining about periods
    # between different between indexes
    for time, fname in local_files.iteritems():
        remote_files.loc[time] = fname
    return remote_files


def list_files(tag='north', sat_id=None, data_path=None, format_str=None):
    """Return a Pandas Series of every file for chosen satellite data

    Parameters
    -----------
    tag : string
        Denotes type of file to load.  Accepted types are 'north' and 'south'.
        (default='north')
    sat_id : string or NoneType
        Specifies the satellite ID for a constellation.  Not used.
        (default=None)
    data_path : string or NoneType
        Path to data directory.  If None is specified, the value previously
        set in Instrument.files.data_path is used.  (default=None)
    format_str : string or NoneType
        User specified file format.  If None is specified, the default
        formats associated with the supplied tags are used. (default=None)

    Returns
    --------
    pysat.Files.from_os : (pysat._files.Files)
        A class containing the verified available files
    """

    if format_str is None and tag is not None:
        if tag == 'north' or tag == 'south':
            hemi_fmt = ''.join(('{year:4d}{month:02d}{day:02d}.', tag,
                                '.grdex'))
            return pysat.Files.from_os(data_path=data_path,
                                       format_str=hemi_fmt)
        else:
            estr = 'Unrecognized tag name for SuperDARN, north or south.'
            raise ValueError(estr)
    elif format_str is None:
        estr = 'A tag name must be passed to SuperDARN.'
        raise ValueError(estr)
    else:
        return pysat.Files.from_os(data_path=data_path, format_str=format_str)


def load(fnames, tag=None, sat_id=None):
    import davitpy
    if len(fnames) <= 0:
        return pysat.DataFrame(None), pysat.Meta(None)
    elif len(fnames) == 1:

        myPtr = davitpy.pydarn.sdio.sdDataPtr(sTime=pysat.datetime(1980, 1, 1),
                                              fileType='grdex',
                                              eTime=pysat.datetime(2250, 1, 1),
                                              hemi=tag,
                                              fileName=fnames[0])
        myPtr.open()

        in_list = []
        in_dict = {'stid': [],
                   'channel': [],
                   'noisemean': [],
                   'noisesd': [],
                   'gsct': [],
                   'nvec': [],
                   'pmax': [],
                   'start_time': [],
                   'end_time': [],
                   'vemax': [],
                   'vemin': [],
                   'pmin': [],
                   'programid': [],
                   'wmax': [],
                   'wmin': [],
                   'freq': []}

        while True:
            info = myPtr.readRec()
            if info is None:
                myPtr.close()
                break

            drift_frame = pds.DataFrame.from_records(info.vector.__dict__,
                                                     nrows=len(info.pmax),
                                                     index=info.vector.index)
            drift_frame['partial'] = 1
            drift_frame.drop('index', axis=1, inplace=True)
            drift_frame.index.name = 'index'
            sum_vec = 0
            for nvec in info.nvec:
                in_list.append(drift_frame.iloc[sum_vec:sum_vec+nvec])
                sum_vec += nvec

            in_dict['stid'].extend(info.stid)
            in_dict['channel'].extend(info.channel)
            in_dict['noisemean'].extend(info.noisemean)
            in_dict['noisesd'].extend(info.noisesd)
            in_dict['gsct'].extend(info.gsct)
            in_dict['nvec'].extend(info.nvec)
            in_dict['pmax'].extend(info.pmax)
            in_dict['start_time'].extend([info.sTime]*len(info.pmax))
            in_dict['end_time'].extend([info.eTime]*len(info.pmax))
            in_dict['vemax'].extend(info.vemax)
            in_dict['vemin'].extend(info.vemin)
            in_dict['pmin'].extend(info.pmin)
            in_dict['programid'].extend(info.programid)
            in_dict['wmax'].extend(info.wmax)
            in_dict['wmin'].extend(info.wmin)
            in_dict['freq'].extend(info.freq)

        output = pds.DataFrame(in_dict)
        output['vector'] = in_list
        output.index = output.start_time
        output.drop('start_time', axis=1, inplace=True)

        return output, pysat.Meta()
    else:
        raise ValueError('Only one filename currently supported.')


# def default(ivm):
#
#    return

def clean(self):
    # remove data when there are no vectors
    idx, = np.where(self['nvec'] > 0)
    self.data = self.data.iloc[idx]

    return


def download(date_array, tag, sat_id, data_path, user=None, password=None):
    """
    Download SuperDARN data from Virginia Tech organized for loading by pysat.

    """

    import warnings

    warnings.warn(" ".join(("Downloads for SuperDARN currently not supported,",
                            "but will be added in a future version.")))



"""
tests the pysat occur_prob object and code
"""

from nose.tools import raises
import warnings
import pysat
from pysat.ssnl import occur_prob


class TestBasics():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        orbit_info = {'index': 'longitude', 'kind': 'longitude'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=orbit_info)
        self.testInst.bounds = (pysat.datetime(2008, 1, 1),
                                pysat.datetime(2008, 1, 31))

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst

    def test_occur_prob_daily_2D_w_bins(self):
        """Runs a basic probability routine daily 2D w/ bins"""
        ans = occur_prob.daily2D(self.testInst, [0, 24, 2], 'slt',
                                 [-60, 60, 3], 'latitude', ['slt'], [12.],
                                 returnBins=True)
        assert abs(ans['slt']['prob'][:, 0] - 0.0).max() < 1.0e-6
        assert abs(ans['slt']['prob'][:, 1] - 1.0).max() < 1.0e-6
        assert (ans['slt']['prob']).shape == (3, 2)
        assert abs(ans['slt']['bin_x'] - [0, 12, 24]).max() < 1.0e-6
        assert abs(ans['slt']['bin_y'] - [-60, -20, 20, 60]).max() < 1.0e-6

    @raises(ValueError)
    def test_occur_prob_daily_2D_w_bad_data_label(self):
        """Catch a data_label that is not list-like"""
        occur_prob.daily2D(self.testInst, [0, 360, 4], 'longitude',
                           [-60, 60, 3], 'latitude', 'slt', [12.])

    @raises(ValueError)
    def test_occur_prob_daily_2D_w_bad_gate(self):
        """Catch a gate that is not list-like"""
        occur_prob.daily2D(self.testInst, [0, 360, 4], 'longitude',
                           [-60, 60, 3], 'latitude', ['slt'], 12.)

    @raises(ValueError)
    def test_occur_prob_daily_2D_w_mismatched_gate_and_data_label(self):
        """Catch a gate that does not match the data_label"""
        occur_prob.daily2D(self.testInst, [0, 360, 4], 'longitude',
                           [-60, 60, 3], 'latitude', ['slt'], [12., 18.])

    def test_occur_prob_by_orbit_2D_w_bins(self):
        """Runs a basic probability routine by orbit 2D"""
        ans = occur_prob.by_orbit2D(self.testInst, [0, 24, 2], 'slt',
                                    [-60, 60, 3], 'latitude', ['slt'], [12.],
                                    returnBins=True)
        assert abs(ans['slt']['prob'][:, 0] - 0.0).max() < 1.0e-6
        assert abs(ans['slt']['prob'][:, 1] - 1.0).max() < 1.0e-6
        assert (ans['slt']['prob']).shape == (3, 2)
        assert abs(ans['slt']['bin_x'] - [0, 12, 24]).max() < 1.0e-6
        assert abs(ans['slt']['bin_y'] - [-60, -20, 20, 60]).max() < 1.0e-6

    def test_occur_prob_daily_3D_w_bins(self):
        """Runs a basic probability routine daily 3D"""
        ans = occur_prob.daily3D(self.testInst, [0, 360, 4], 'longitude',
                                 [-60, 60, 3], 'latitude', [0, 24, 2], 'slt',
                                 ['slt'], [12.], returnBins=True)
        assert abs(ans['slt']['prob'][0, :, :] - 0.0).max() < 1.0e-6
        assert abs(ans['slt']['prob'][-1, :, :] - 1.0).max() < 1.0e-6
        assert (ans['slt']['prob']).shape == (2, 3, 4)
        assert abs(ans['slt']['bin_x'] - [0, 90, 180, 270, 360]).max() < 1.0e-6
        assert abs(ans['slt']['bin_y'] - [-60, -20, 20, 60]).max() < 1.0e-6
        assert abs(ans['slt']['bin_z'] - [0, 12, 24]).max() < 1.0e-6

    @raises(ValueError)
    def test_occur_prob_daily_3D_w_bad_data_label(self):
        """Catch a data_label that is not list-like"""
        occur_prob.daily3D(self.testInst, [0, 360, 4], 'longitude',
                           [-60, 60, 3], 'latitude', [0, 24, 2], 'slt',
                           'slt', [12.])

    @raises(ValueError)
    def test_occur_prob_daily_3D_w_bad_gate(self):
        """Catch a gate that is not list-like"""
        occur_prob.daily3D(self.testInst, [0, 360, 4], 'longitude',
                           [-60, 60, 3], 'latitude', [0, 24, 2], 'slt',
                           ['slt'], 12.)

    @raises(ValueError)
    def test_occur_prob_daily_3D_w_mismatched_gate_and_data_label(self):
        """Catch a gate that does not match the data_label"""
        occur_prob.daily3D(self.testInst, [0, 360, 4], 'longitude',
                           [-60, 60, 3], 'latitude', [0, 24, 2], 'slt',
                           ['slt'], [12., 18.])

    def test_occur_prob_by_orbit_3D_w_bins(self):
        """Runs a basic probability routine by orbit 3D"""
        ans = occur_prob.by_orbit3D(self.testInst, [0, 360, 4], 'longitude',
                                    [-60, 60, 3], 'latitude',
                                    [0, 24, 2], 'slt', ['slt'], [12.],
                                    returnBins=True)
        assert abs(ans['slt']['prob'][0, :, :] - 0.0).max() < 1.0e-6
        assert abs(ans['slt']['prob'][-1, :, :] - 1.0).max() < 1.0e-6
        assert (ans['slt']['prob']).shape == (2, 3, 4)
        assert abs(ans['slt']['bin_x'] - [0, 90, 180, 270, 360]).max() < 1.0e-6
        assert abs(ans['slt']['bin_y'] - [-60, -20, 20, 60]).max() < 1.0e-6
        assert abs(ans['slt']['bin_z'] - [0, 12, 24]).max() < 1.0e-6


class TestDeprecation():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        warnings.simplefilter("always")

    def teardown(self):
        """Runs after every method to clean up previous testing."""

    def test_deprecation_warning_daily_2D(self):
        """Test if occur_prob.daily2D is deprecated"""

        with warnings.catch_warnings(record=True) as war:
            try:
                occur_prob.daily2D(None, [0, 24, 2], 'slt',
                                   [-60, 60, 3], 'latitude', ['slt'], [12.])
            except TypeError:
                # Setting inst to None should produce a TypeError after
                # warning is generated
                pass

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning

    def test_deprecation_warning_by_orbit_2D(self):
        """Test if occur_prob.by_orbit2D is deprecated"""

        with warnings.catch_warnings(record=True) as war:
            try:
                occur_prob.by_orbit2D(None, [0, 24, 2], 'slt',
                                      [-60, 60, 3], 'latitude', ['slt'], [12.])
            except AttributeError:
                # Setting inst to None should produce a AttributeError after
                # warning is generated
                pass

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning

    def test_deprecation_warning_daily_3D(self):
        """Test if occur_prob.daily3D is deprecated"""

        with warnings.catch_warnings(record=True) as war:
            try:
                occur_prob.daily3D(None, [0, 360, 4], 'longitude',
                                   [-60, 60, 3], 'latitude', [0, 24, 2], 'slt',
                                   ['slt'], [12.])
            except TypeError:
                # Setting inst to None should produce a TypeError after
                # warning is generated
                pass

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning

    def test_deprecation_warning_by_orbit_3D(self):
        """Test if occur_prob.by_orbit3D is deprecated"""

        with warnings.catch_warnings(record=True) as war:
            try:
                occur_prob.by_orbit3D(None, [0, 24, 2], 'slt',
                                      [-60, 60, 3], 'latitude', [0, 360, 4],
                                      'longitude', ['slt'], [12.])
            except AttributeError:
                # Setting inst to None should produce a AttributeError after
                # warning is generated
                pass

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning



"""
Create a constellation with one testing instrument

Attributes
----------
instruments : list
    List of pysat.Instrument objects

"""

import pysat

instruments = [pysat.Instrument('pysat', 'testing', clean_level='clean',
                                update_files=True)]



#!/usr/bin/env python
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.1199703
# ----------------------------------------------------------------------------
"""
pysat user module registry utilities

This module allows pysat to provide direct access to external or custom
instrument modules by maintaining information about these instrument modules.

Examples
--------
Instrument support modules must be registered before use. This may be done
individually or for a collection of Instruments at once. For example, assume
there is an implementation for `myInstrument` in the module
`my.package.myInstrument` with platform and name attributes 'myplatform' and
'myname'. Such an instrument may be registered with
::

    registry.register('my.package.myInstrument')

The full module name "my.package.myInstrument" will be registered in
pysat.params['user_modules'] and stored as a dict of dicts keyed by platform
and name.

Once registered, subsequent calls to Instrument may use the platform and name
string identifiers.
::

    Instrument('myplatform', 'myname')

A full suite of instrument support modules may be registered at once using
::

    # General form where my.package contains a collection of
    # submodules to support Instrument data sets.
    registry.register_by_module(my.package)

    # Register published packages from pysat team
    import pysatSpaceWeather
    registry.register_by_module(pysatSpaceWeather.instruments)

    import pysatNASA
    registry.register_by_module(pysatNASA.instruments)

    import pysatModels
    registry.register_by_module(pysatModels.models)

"""

import importlib
import logging

import pysat
import pysat.tests.instrument_test_class as itc
logger = logging.getLogger(__name__)


def load_saved_modules():
    """Load registered pysat.Instrument modules

    Returns
    -------
    dict of dicts
        instrument module strings are keyed by
        platform then name

    """

    return pysat.params['user_modules']


def store():
    """Store current registry onto disk"""

    pysat.params.store()

    return


def register(module_names, overwrite=False):
    """Registers a user pysat.Instrument module by name

    Enables instantiation of a third-party Instrument
    module using
    ::

        inst = pysat.Instrument(platform, name, tag=tag, inst_id=inst_id)

    Parameters
    ----------
    module_names : list-like of str
        specify package name and instrument modules
    overwrite : bool
        If True, an existing registration will be updated
        with the new module information.

    Raises
    ------
    ValueError
        If a new module is input with a platform and name that is already
        associated with a registered module and the overwrite flag is set to
        False.

    Warnings
    --------
    Registering a module that contains code other than pysat instrument
    files could result in unexpected consequences.

    Note
    ----
    Modules should be importable using
    ::

        from my.package.name import my_instrument

    Module names do not have to follow the pysat platform_name naming
    convection.

    Current registered modules bay be found at
    ::

        pysat.params['user_modules']

    which is stored as a dict of dicts keyed by platform and name.

    Examples
    --------
    ::

        from pysat import Instrument
        from pysat.utils import registry

        registry.register('my.package.name.myInstrument')

        testInst = Instrument(platform, name)

    """

    for mod_name in module_names:
        # First, ensure module string directs to something importable
        try:
            inst_module = importlib.import_module(mod_name)
        except Exception:
            # Log then preserve trace and propagate error
            estr = ' '.join(('There was a problem trying to import', mod_name))
            logger.error(estr)
            raise

        # Second, check that module is itself pysat compatible
        validate = itc.InstTestClass()

        # Work with test code, create dummy structure to make things work
        class Foo(object):
            pass
        validate.inst_loc = Foo()

        # Parse string to get package part and instrument module part
        parse = mod_name.split('.')

        # Module name without package
        mod_part = parse[-1]

        # The package preamble
        pack_part = parse[:-1]

        # Assign package info to Test class
        validate.inst_loc.__name__ = '.'.join(pack_part)

        # Run tests
        validate.test_modules_standard(mod_part)
        validate.test_standard_function_presence(mod_part)

        # Registry is a dict of dicts with platform, name, and module string.
        # Get the platform and name identifiers from imported module
        platform = inst_module.platform
        name = inst_module.name

        # Only register module if not already present. Multiple names are
        # allowed for a single platform
        if platform not in pysat.params['user_modules']:
            # setup `of dict` part of dict of dicts
            pysat.params.data['user_modules'][platform] = {}

        # Only register name if it is not present under platform
        if name not in pysat.params['user_modules'][platform]:
            logger.info('Registering user module {}'.format(mod_name))
            # Add to current user modules structure and store it to disk
            pysat.params.data['user_modules'][platform][name] = mod_name
            store()
        else:
            # Platform/name combination already registered. Check to see if
            # this is a new package or just a redundant assignment
            if mod_name != pysat.params['user_modules'][platform][name]:
                # New assignment, check for overwrite flag
                if not overwrite:
                    estr = ' '.join(('An instrument has already been ',
                                     'registered for platform:', platform,
                                     'and name:', name,
                                     'which maps to:', mod_name, 'To assign',
                                     'a new module the overwrite flag',
                                     'must be enabled.'))
                    raise ValueError(estr)
                else:
                    # Overwrite with new module information
                    pysat.params.data['user_modules'][platform][name] = mod_name
                    store()

    return


def register_by_module(module):
    """Register all sub-modules attached to input module

    Enables instantiation of a third-party Instrument module using
    ::

        inst = pysat.Instrument(platform, name)

    Parameters
    ----------
    module : Python module
        Module with one or more pysat.Instrument support modules
        attached as sub-modules to the input `module`

    Raises
    ------
    ValueError
        If platform and name associated with a module are already registered

    Note
    ----
    Gets a list of sub-modules by using the `__all__` attribute,
    defined in the module's `__init__.py`

    Examples
    --------
    ::

        import pysat
        import pysatModels
        pysat.utils.registry.register_by_module(pysatModels.models)

    """

    # Get a list of all user specified modules attached to module
    module_names = module.__all__

    # Add in package preamble
    module_names = [module.__name__ + '.' + mod for mod in module_names]

    # Register all of the sub-modules
    register(module_names)

    return


def remove(platforms, names):
    """Removes module from registered user modules

    Parameters
    ----------
    platforms : list-like of str
        Platform identifiers to remove
    names : list-like of str
        Name identifiers, paired with platforms, to remove.
        If the names element paired with the platform element is None,
        then all instruments under the specified platform will be
        removed. Should be the same type as `platforms`.

    Raises
    ------
    ValueError
        If platform and/or name are not currently registered

    Note
    ----
    Current registered user modules available at pysat.params['user_modules']

    Examples
    --------
    ::

        platforms = ['platform1', 'platform2']
        names = ['name1', 'name2']

        # remove all instruments with platform=='platform1'
        registry.remove(['platform1'], [None])
        # remove all instruments with platform 'platform1' or 'platform2'
        registry.remove(platforms, [None, None])
        # remove all instruments with 'platform1', and individual instrument
        # for 'platform2', 'name2'
        registry.remove(platforms, [None, 'name2']
        # remove 'platform1', 'name1', as well as 'platform2', 'name2'
        registry.remove(platforms, names)

    """

    # ensure equal number of inputs
    if len(names) != len(platforms):
        estr = "".join(("The number of 'platforms' and 'names' must be the ",
                        "same, or names must be None, which removes all ",
                        "instruments under each platform."))
        raise ValueError(estr)

    # iterate over inputs and remove modules
    for platform, name in zip(platforms, names):
        if platform in pysat.params['user_modules']:
            if name is None:
                # remove platform entirely
                pysat.params['user_modules'].pop(platform)
                # store
                store()
            else:
                # name supplied, remove single module
                if name in pysat.params['user_modules'][platform]:
                    # remove module
                    pysat.params['user_modules'][platform].pop(name)
                else:
                    # name not in platform
                    estr = ''.join((platform, ', ', name, ': not a registered ',
                                    'instrument module.'))
                    logger.info(estr)
                # remove platform if no remaining instruments
                if len(pysat.params['user_modules'][platform]) == 0:
                    pysat.params['user_modules'].pop(platform)
                # store
                store()
        else:
            # info string if module not registered
            estr = ''.join((platform, ': is not a registered ',
                            'instrument platform.'))
            # platform not in registered modules
            logger.info(estr)

    return



import pysat

"""
Creates an instrument used to test the constellation class's addition function.

"""

instruments = [pysat.Instrument('pysat', 'testing', tag='ascend',
                                clean_level='clean'),
               pysat.Instrument('pysat', 'testing', tag='plus10',
                                clean_level='clean')]



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import absolute_import
"""
Instrument independent seasonal averaging routine. Supports averaging
1D and 2D data.
"""

import pysat
import numpy as np
import pandas as pds
import collections
import warnings


def median1D(const, bin1, label1, data_label, auto_bin=True, returnData=False):
    """Return a 1D median of data_label over a season and label1

    .. deprecated:: 2.2.0
      `median1D` will be removed in pysat 3.0.0, it will
      be added to pysatSeasons

    Parameters
    ----------
    const: Constellation or Instrument
        Constellation or Instrument object
    bin1: (array-like)
        List holding [min, max, number of bins] or array-like containing
        bin edges
    label1: (string)
        data column name that the binning will be performed over (i.e., lat)
    data_label: (list-like )
        contains strings identifying data product(s) to be averaged
    auto_bin: if True, function will create bins from the min, max and
              number of bins. If false, bin edges must be manually entered
    returnData : (boolean)
        Return data in output dictionary as well as statistics

    Returns
    -------
    median : dictionary
        1D median accessed by data_label as a function of label1
        over the season delineated by bounds of passed instrument objects.
        Also includes 'count' and 'avg_abs_dev' as well as the values of
        the bin edges in 'bin_x'

    """

    warnings.warn(' '.join(["This function is deprecated here and will be",
                            "removed in pysat 3.0.0. Please use",
                            "pysatSeasons instead:"
                            "https://github.com/pysat/pysatSeasons"]),
                  DeprecationWarning, stacklevel=2)

    # const is either an Instrument or a Constellation, and we want to
    #  iterate over it.
    # If it's a Constellation, then we can do that as is, but if it's
    #  an Instrument, we just have to put that Instrument into something
    #  that will yeild that Instrument, like a list.
    if isinstance(const, pysat.Instrument):
        const = [const]
    elif not isinstance(const, pysat.Constellation):
        raise ValueError("Parameter must be an Instrument or a Constellation.")

    # create bins
    # seems to create the boundaries used for sorting into bins
    if auto_bin:
        binx = np.linspace(bin1[0], bin1[1], bin1[2]+1)
    else:
        binx = np.array(bin1)

    # how many bins are used
    numx = len(binx)-1
    # how many different data products
    numz = len(data_label)

    # create array to store all values before taking median
    # the indices of the bins/data products? used for looping.
    xarr = np.arange(numx)
    zarr = np.arange(numz)
    # 3d array:  stores the data that is sorted into each bin? - in a deque
    ans = [[collections.deque() for i in xarr] for k in zarr]

    for inst in const:
        # do loop to iterate over instrument season
        # probably iterates by date but that all depends on the
        # configuration of that particular instrument.
        # either way, it iterates over the instrument, loading successive
        # data between start and end bounds
        for inst in inst:
            # collect data in bins for averaging
            if len(inst.data) != 0:
                # sort the data into bins (x) based on label 1
                # (stores bin indexes in xind)
                xind = np.digitize(inst.data[label1], binx) - 1
                # for each possible x index
                for xi in xarr:
                    # get the indicies of those pieces of data in that bin
                    xindex, = np.where(xind == xi)
                    if len(xindex) > 0:
                        # for each data product label zk
                        for zk in zarr:
                            # take the data (already filtered by x), select the
                            # data, put it in a list, and extend the deque
                            idata = inst.data.iloc[xindex]
                            ans[zk][xi].extend(idata[data_label[zk]].tolist())

    # Calculate the 1D median
    return _calc_1d_median(ans, data_label, binx, xarr, zarr, numx, numz,
                           returnData)


def median2D(const, bin1, label1, bin2, label2, data_label,
             returnData=False, auto_bin=True):
    """Return a 2D average of data_label over a season and label1, label2.

    .. deprecated:: 2.2.0
      `median2D` will be removed in pysat 3.0.0, it will
      be added to pysatSeasons

    Parameters
    ----------
        const: Constellation or Instrument
        bin#: [min, max, number of bins], or array-like containing bin edges
        label#: string
            identifies data product for bin#
        data_label: list-like
            contains strings identifying data product(s) to be averaged
        auto_bin: if True, function will create bins from the min, max and
                  number of bins. If false, bin edges must be manually entered

    Returns
    -------
    median : dictionary
        2D median accessed by data_label as a function of label1 and label2
        over the season delineated by bounds of passed instrument objects.
        Also includes 'count' and 'avg_abs_dev' as well as the values of
        the bin edges in 'bin_x' and 'bin_y'.

    """

    warnings.warn(' '.join(["This function is deprecated here and will be",
                            "removed in pysat 3.0.0. Please use",
                            "pysatSeasons instead:"
                            "https://github.com/pysat/pysatSeasons"]),
                  DeprecationWarning, stacklevel=2)

    # const is either an Instrument or a Constellation, and we want to
    #  iterate over it.
    # If it's a Constellation, then we can do that as is, but if it's
    #  an Instrument, we just have to put that Instrument into something
    #  that will yeild that Instrument, like a list.
    if isinstance(const, pysat.Instrument):
        const = [const]
    elif not isinstance(const, pysat.Constellation):
        raise ValueError("Parameter must be an Instrument or a Constellation.")

    # create bins
    # seems to create the boundaries used for sorting into bins
    if auto_bin:
        binx = np.linspace(bin1[0], bin1[1], bin1[2]+1)
        biny = np.linspace(bin2[0], bin2[1], bin2[2]+1)
    else:
        binx = np.array(bin1)
        biny = np.array(bin2)

    # how many bins are used
    numx = len(binx)-1
    numy = len(biny)-1
    # how many different data products
    numz = len(data_label)

    # create array to store all values before taking median
    # the indices of the bins/data products? used for looping.
    yarr = np.arange(numy)
    xarr = np.arange(numx)
    zarr = np.arange(numz)
    # 3d array:  stores the data that is sorted into each bin? - in a deque
    ans = [[[collections.deque() for i in xarr] for j in yarr] for k in zarr]

    for inst in const:
        # do loop to iterate over instrument season
        # probably iterates by date but that all depends on the
        # configuration of that particular instrument.
        # either way, it iterates over the instrument, loading successive
        # data between start and end bounds
        for inst in inst:
            # collect data in bins for averaging
            if len(inst.data) != 0:
                # sort the data into bins (x) based on label 1
                # (stores bin indexes in xind)
                xind = np.digitize(inst.data[label1], binx)-1
                # for each possible x index
                for xi in xarr:
                    # get the indicies of those pieces of data in that bin
                    xindex, = np.where(xind == xi)
                    if len(xindex) > 0:
                        # look up the data along y (label2) at that set of
                        # indices (a given x)
                        yData = inst.data.iloc[xindex]
                        # digitize that, to sort data into bins along y
                        # (label2) (get bin indexes)
                        yind = np.digitize(yData[label2], biny)-1
                        # for each possible y index
                        for yj in yarr:
                            # select data with this y index (and we already
                            # filtered for this x index)
                            yindex, = np.where(yind == yj)
                            if len(yindex) > 0:
                                # for each data product label zk
                                for zk in zarr:
                                    # take the data (already filtered by x)
                                    # filter it by y and
                                    # select the data product, put it in a
                                    # list, and extend the deque
                                    indlab = yData.columns.get_loc(data_label[zk])
                                    ans[zk][yj][xi].extend(yData.iloc[yindex,
                                                                      indlab].tolist())

    return _calc_2d_median(ans, data_label, binx, biny, xarr, yarr, zarr,
                           numx, numy, numz, returnData)


def _calc_2d_median(ans, data_label, binx, biny, xarr, yarr, zarr, numx,
                    numy, numz, returnData=False):

    # set up output arrays
    medianAns = [[[None for i in xarr] for j in yarr] for k in zarr]
    countAns = [[[None for i in xarr] for j in yarr] for k in zarr]
    devAns = [[[None for i in xarr] for j in yarr] for k in zarr]

    # all of the loading and storing data is done
    # determine what kind of data is stored
    # if just numbers, then use numpy arrays to store data
    # if the data is a more generalized object, use lists to store data
    # need to find first bin with data
    dataType = [None for i in np.arange(numz)]
    # for each data product label, find the first nonempty bin
    # and select its type
    for zk in zarr:
        breakNow = False
        for yj in yarr:
            for xi in xarr:
                if len(ans[zk][yj][xi]) > 0:
                    dataType[zk] = type(ans[zk][yj][xi][0])
                    breakNow = True
                    break
            if breakNow:
                break

    # determine if normal number objects are being used or if there
    # are more complicated objects
    objArray = [False] * len(zarr)
    for i, thing in enumerate(dataType):
        if thing == pds.core.series.Series:
            objArray[i] = 'S'
        elif thing == pds.core.frame.DataFrame:
            objArray[i] = 'F'
        else:
            # other, simple scalaRs
            objArray[i] = 'R'

    objArray = np.array(objArray)

    # if some pandas data series are returned in average, return a list
    objidx, = np.where(objArray == 'S')
    if len(objidx) > 0:
        for zk in zarr[objidx]:
            for yj in yarr:
                for xi in xarr:
                    if len(ans[zk][yj][xi]) > 0:
                        ans[zk][yj][xi] = list(ans[zk][yj][xi])
                        medianAns[zk][yj][xi] = \
                            pds.DataFrame(ans[zk][yj][xi]).median(axis=0)
                        countAns[zk][yj][xi] = len(ans[zk][yj][xi])
                        devAns[zk][yj][xi] = \
                            pds.DataFrame([abs(temp - medianAns[zk][yj][xi])
                                           for temp in
                                           ans[zk][yj][xi]]).median(axis=0)

    # if some pandas DataFrames are returned in average, return a list
    objidx, = np.where(objArray == 'F')
    if len(objidx) > 0:
        for zk in zarr[objidx]:
            for yj in yarr:
                for xi in xarr:
                    if len(ans[zk][yj][xi]) > 0:
                        ans[zk][yj][xi] = list(ans[zk][yj][xi])
                        countAns[zk][yj][xi] = len(ans[zk][yj][xi])
                        test = pds.Panel.from_dict(dict([(i, temp)
                                                   for i, temp in
                                                   enumerate(ans[zk][yj][xi])]))
                        medianAns[zk][yj][xi] = test.median(axis=0)
                        devAns[zk][yj][xi] = \
                            (test.subtract(medianAns[zk][yj][xi],
                                           axis=0)).abs().median(axis=0,
                                                                 skipna=True)

    objidx, = np.where(objArray == 'R')
    if len(objidx) > 0:
        for zk in zarr[objidx]:
            medianAns[zk] = np.zeros((numy, numx))*np.nan
            countAns[zk] = np.zeros((numy, numx))*np.nan
            devAns[zk] = np.zeros((numy, numx))*np.nan
            for yj in yarr:
                for xi in xarr:
                    # convert deque storing data into numpy array
                    ans[zk][yj][xi] = np.array(ans[zk][yj][xi])
                    # filter out an NaNs in the arrays
                    idx, = np.where(np.isfinite(ans[zk][yj][xi]))
                    ans[zk][yj][xi] = (ans[zk][yj][xi])[idx]
                    # perform median averaging
                    if len(idx) > 0:
                        medianAns[zk][yj, xi] = np.median(ans[zk][yj][xi])
                        countAns[zk][yj, xi] = len(ans[zk][yj][xi])
                        devAns[zk][yj, xi] = np.median(abs(ans[zk][yj][xi] -
                                                       medianAns[zk][yj, xi]))

    # prepare output
    output = {}
    for i, label in enumerate(data_label):
        output[label] = {'median': medianAns[i],
                         'count': countAns[i],
                         'avg_abs_dev': devAns[i],
                         'bin_x': binx,
                         'bin_y': biny}

        if returnData:
            output[label]['data'] = ans[i]

    return output


# simple averaging through multiple iterations

def mean_by_day(inst, data_label):
    """Mean of data_label by day over Instrument.bounds

    .. deprecated:: 2.2.0
      `mean_by_day` will be removed in pysat 3.0.0, it will
      be added to pysatSeasons

    Parameters
    ----------
    data_label : string
        string identifying data product to be averaged

    Returns
    -------
    mean : pandas Series
        simple mean of data_label indexed by day

    """

    warnings.warn(' '.join(["This function is deprecated here and will be",
                            "removed in pysat 3.0.0. Please use",
                            "pysatSeasons instead:"
                            "https://github.com/pysat/pysatSeasons"]),
                  DeprecationWarning, stacklevel=2)

    return _core_mean(inst, data_label, by_day=True)


def mean_by_orbit(inst, data_label):
    """Mean of data_label by orbit over Instrument.bounds

    .. deprecated:: 2.2.0
      `mean_by_orbit` will be removed in pysat 3.0.0, it will
      be added to pysatSeasons

    Parameters
    ----------
    data_label : string
        string identifying data product to be averaged

    Returns
    -------
    mean : pandas Series
        simple mean of data_label indexed by start of each orbit

    """

    warnings.warn(' '.join(["This function is deprecated here and will be",
                            "removed in pysat 3.0.0. Please use",
                            "pysatSeasons instead:"
                            "https://github.com/pysat/pysatSeasons"]),
                  DeprecationWarning, stacklevel=2)

    return _core_mean(inst, data_label, by_orbit=True)


def mean_by_file(inst, data_label):
    """Mean of data_label by orbit over Instrument.bounds

    .. deprecated:: 2.2.0
      `mean_by_file` will be removed in pysat 3.0.0, it will
      be added to pysatSeasons

    Parameters
    ----------
    data_label : string
        string identifying data product to be averaged

    Returns
    -------
    mean : pandas Series
        simple mean of data_label indexed by start of each file

    """

    warnings.warn(' '.join(["This function is deprecated here and will be",
                            "removed in pysat 3.0.0. Please use",
                            "pysatSeasons instead:"
                            "https://github.com/pysat/pysatSeasons"]),
                  DeprecationWarning, stacklevel=2)

    return _core_mean(inst, data_label, by_file=True)


def _core_mean(inst, data_label, by_orbit=False, by_day=False, by_file=False):

    if by_orbit:
        iterator = inst.orbits
    elif by_day or by_file:
        iterator = inst
    else:
        raise ValueError('A choice must be made, by day, file, or orbit')

    # create empty series to hold result
    mean_val = pds.Series()
    # iterate over season, calculate the mean
    for inst in iterator:
        if not inst.data.empty:
            # compute mean absolute using pandas functions and store
            # data could be an image, or lower dimension, account for 2D
            # and lower
            data = inst[data_label]
            data.dropna(inplace=True)

            if by_orbit or by_file:
                date = inst.data.index[0]
            else:
                date = inst.date
            # perform average
            mean_val[date] = \
                pysat.ssnl.computational_form(data).mean(axis=0,
                                                         skipna=True)

    del iterator
    return mean_val


def _calc_1d_median(ans, data_label, binx, xarr, zarr, numx, numz,
                    returnData=False):
    """Calculate the 1D median

    Parameters
    ----------

    Returns
    ------

    Notes
    -----
    This is an overcomplicated way of doing this.  Try and simplify later

    """

    # set up output arrays
    medianAns = [[None for i in xarr] for k in zarr]
    countAns = [[None for i in xarr] for k in zarr]
    devAns = [[None for i in xarr] for k in zarr]

    # all of the loading and storing data is done
    # determine what kind of data is stored
    # if just numbers, then use numpy arrays to store data
    # if the data is a more generalized object, use lists to store data
    # need to find first bin with data
    dataType = [None for i in np.arange(numz)]
    # for each data product label, find the first nonempty bin
    # and select its type
    for zk in zarr:
        for xi in xarr:
            if len(ans[zk][xi]) > 0:
                dataType[zk] = type(ans[zk][xi][0])
                break

    # determine if normal number objects are being used or if there
    # are more complicated objects
    objArray = [False] * len(zarr)
    for i, thing in enumerate(dataType):
        if thing == pds.core.series.Series:
            objArray[i] = 'S'
        elif thing == pds.core.frame.DataFrame:
            objArray[i] = 'F'
        else:
            # other, simple scalaRs
            objArray[i] = 'R'

    objArray = np.array(objArray)

    # if some pandas data series are returned in average, return a list
    objidx, = np.where(objArray == 'S')
    if len(objidx) > 0:
        for zk in zarr[objidx]:
            for xi in xarr:
                if len(ans[zk][xi]) > 0:
                    ans[zk][xi] = list(ans[zk][xi])
                    medianAns[zk][xi] = pds.DataFrame(ans[zk][xi]).median(axis=0)
                    countAns[zk][xi] = len(ans[zk][xi])
                    devAns[zk][xi] = pds.DataFrame([abs(temp
                                                        - medianAns[zk][xi]) \
                                    for temp in ans[zk][xi] ]).median(axis=0)

    # if some pandas DataFrames are returned in average, return a list
    objidx, = np.where(objArray == 'F')
    if len(objidx) > 0:
        for zk in zarr[objidx]:
            for xi in xarr:
                if len(ans[zk][xi]) > 0:
                    ans[zk][xi] = list(ans[zk][xi])
                    countAns[zk][xi] = len(ans[zk][xi])
                    test = pds.Panel.from_dict(dict([(i, temp) for i, temp \
                                                    in enumerate(ans[zk][xi])]))
                    medianAns[zk][xi] = test.median(axis=0)
                    devAns[zk][xi] = \
                        (test.subtract(medianAns[zk][xi],
                                       axis=0)).abs().median(axis=0,
                                                             skipna=True)

    objidx, = np.where(objArray == 'R')
    if len(objidx) > 0:
        for zk in zarr[objidx]:
            medianAns[zk] = np.full(numx, fill_value=np.nan)
            countAns[zk] = np.full(numx, fill_value=np.nan)
            devAns[zk] = np.full(numx, fill_value=np.nan)
            for xi in xarr:
                # convert deque storing data into numpy array
                ans[zk][xi] = np.array(ans[zk][xi])
                # filter out an NaNs in the arrays
                idx, = np.where(np.isfinite(ans[zk][xi]))
                ans[zk][xi] = (ans[zk][xi])[idx]
                # perform median averaging
                if len(idx) > 0:
                    medianAns[zk][xi] = np.median(ans[zk][xi])
                    countAns[zk][xi] = len(ans[zk][xi])
                    devAns[zk][xi] = np.median(abs(ans[zk][xi] -
                                                   medianAns[zk][xi]))

    # prepare output
    output = {}
    for i, label in enumerate(data_label):
        output[label] = {'median': medianAns[i],
                         'count': countAns[i],
                         'avg_abs_dev': devAns[i],
                         'bin_x': binx}

        if returnData:
            output[label]['data'] = ans[i]

    return output



# -*- coding: utf-8 -*-
"""
Produces fake instrument data for testing.
"""
from __future__ import print_function
from __future__ import absolute_import
import functools
import numpy as np

import pandas as pds

import pysat
from pysat.instruments.methods import testing as mm_test

# pysat required parameters
platform = 'pysat'
name = 'testing'

# dictionary of data 'tags' and corresponding description
# tags are used to choose the behaviour of dummy1
tags = {'': 'Regular testing data set',
        'ascend': 'Ascending Integers from 0 testing data set',
        'descend': 'Descending Integers from 0 testing data set',
        'plus10': 'Ascending Integers from 10 testing data set',
        'fives': 'All 5s testing data set',
        'mlt_offset': 'dummy1 is offset by five from regular testing set'}

# dictionary of satellite IDs, list of corresponding tags
# a numeric string can be used in sat_id to change the number of points per day
sat_ids = {'': ['', 'ascend', 'descend', 'plus10', 'fives', 'mlt_offset']}
_test_dates = {'': {'': pysat.datetime(2009, 1, 1)}}


def init(self):
    """Initializes the Instrument object with instrument specific values.

    Runs once upon instantiation.

    Shifts time index of files by 5-minutes if mangle_file_dates
    set to True at pysat.Instrument instantiation.

    Creates a file list for a given range if the file_date_range
    keyword is set at instantiation.

    Parameters
    ----------
    inst : pysat.Instrument
        This object
    file_date_range : pds.date_range
        Optional keyword argument that specifies the range of dates for which
        test files will be created
    mangle_file_dates : bool
        If True, the loaded file list time index is shifted by 5-minutes.

    """

    self.new_thing = True

    # work on file index if keyword present
    if self.kwargs['file_date_range'] is not None:
        # set list files routine to desired date range
        # attach to the instrument object
        fdr = self.kwargs['file_date_range']
        self._list_rtn = functools.partial(list_files, file_date_range=fdr)
        self.files.refresh()

    # mess with file dates if kwarg option present
    if self.kwargs['mangle_file_dates']:
        self.files.files.index = \
            self.files.files.index + pds.DateOffset(minutes=5)


def default(self):
    """Default customization function.

    This routine is automatically applied to the Instrument object
    on every load by the pysat nanokernel (first in queue).

    Parameters
    ----------
    self : pysat.Instrument
        This object

    """

    pass


def load(fnames, tag=None, sat_id=None, sim_multi_file_right=False,
         sim_multi_file_left=False, root_date=None, file_date_range=None,
         malformed_index=False, mangle_file_dates=False):
    """ Loads the test files

    Parameters
    ----------
    fnames : list
        List of filenames
    tag : str or NoneType
        Instrument tag (accepts '')
    sat_id : str or NoneType
        Instrument satellite ID (accepts '' or a number (i.e., '10'), which
        specifies the number of data points to include in the test instrument)
    sim_multi_file_right : boolean
        Adjusts date range to be 12 hours in the future or twelve hours beyond
        root_date (default=False)
    sim_multi_file_left : boolean
        Adjusts date range to be 12 hours in the past or twelve hours before
        root_date (default=False)
    root_date : NoneType
        Optional central date, uses _test_dates if not specified.
        (default=None)
    file_date_range : pds.date_range or NoneType
        Range of dates for files or None, if this optional arguement is not
        used. Shift actually performed by the init function.
        (default=None)
    malformed_index : boolean
        If True, time index will be non-unique and non-monotonic.
    mangle_file_dates : bool
        If True, the loaded file list time index is shifted by 5-minutes.
        This shift is actually performed by the init function.

    Returns
    -------
    data : (pds.DataFrame)
        Testing data
    meta : (pysat.Meta)
        Metadataxs

    """

    # create an artifical satellite data set
    iperiod = mm_test.define_period()
    drange = mm_test.define_range()
    uts, index, date = mm_test.generate_times(fnames, sat_id, freq='1S')

    # Specify the date tag locally and determine the desired date range
    pds_offset = pds.DateOffset(hours=12)
    if sim_multi_file_right:
        root_date = root_date or _test_dates[''][''] + pds_offset
    elif sim_multi_file_left:
        root_date = root_date or _test_dates[''][''] - pds_offset
    else:
        root_date = root_date or _test_dates['']['']

    data = pysat.DataFrame(uts, columns=['uts'])

    # need to create simple orbits here. Have start of first orbit default
    # to 1 Jan 2009, 00:00 UT. 14.84 orbits per day
    time_delta = date - root_date
    data['mlt'] = mm_test.generate_fake_data(time_delta.total_seconds(),
                                             uts, period=iperiod['lt'],
                                             data_range=drange['lt'])

    # do slt, 20 second offset from mlt
    data['slt'] = mm_test.generate_fake_data(time_delta.total_seconds()+20,
                                             uts, period=iperiod['lt'],
                                             data_range=drange['lt'])

    # create a fake longitude, resets every 6240 seconds
    # sat moves at 360/5820 deg/s, Earth rotates at 360/86400, takes extra time
    # to go around full longitude
    data['longitude'] = mm_test.generate_fake_data(time_delta.total_seconds(),
                                                   uts, period=iperiod['lon'],
                                                   data_range=drange['lon'])

    # create latitude area for testing polar orbits
    angle = mm_test.generate_fake_data(time_delta.total_seconds(),
                                       uts, period=iperiod['angle'],
                                       data_range=drange['angle'])
    data['latitude'] = 90.0 * np.cos(angle)

    # fake orbit number
    fake_delta = date - (_test_dates[''][''] - pds.DateOffset(years=1))
    data['orbit_num'] = mm_test.generate_fake_data(fake_delta.total_seconds(),
                                                   uts, period=iperiod['lt'],
                                                   cyclic=False)

    # create some fake data to support testing of averaging routines
    mlt_int = data['mlt'].astype(int)
    long_int = (data['longitude'] / 15.0).astype(int)
    if tag == 'ascend':
        data['dummy1'] = [i for i in range(len(data['mlt']))]
    elif tag == 'descend':
        data['dummy1'] = [-i for i in range(len(data['mlt']))]
    elif tag == 'plus10':
        data['dummy1'] = [i + 10 for i in range(len(data['mlt']))]
    elif tag == 'fives':
        data['dummy1'] = [5 for i in range(len(data['mlt']))]
    elif tag == 'mlt_offset':
        data['dummy1'] = mlt_int + 5
    else:
        data['dummy1'] = mlt_int
    data['dummy2'] = long_int
    data['dummy3'] = mlt_int + long_int * 1000.0
    data['dummy4'] = uts
    data['string_dummy'] = ['test'] * len(data)
    data['unicode_dummy'] = [u'test'] * len(data)
    data['int8_dummy'] = np.ones(len(data), dtype=np.int8)
    data['int16_dummy'] = np.ones(len(data), dtype=np.int16)
    data['int32_dummy'] = np.ones(len(data), dtype=np.int32)
    data['int64_dummy'] = np.ones(len(data), dtype=np.int64)

    if malformed_index:
        index = index.tolist()
        # nonmonotonic
        index[0:3], index[3:6] = index[3:6], index[0:3]
        # non unique
        index[6:9] = [index[6]]*3

    data.index = index
    data.index.name = 'Epoch'
    return data, meta.copy()


list_files = functools.partial(mm_test.list_files, test_dates=_test_dates)
download = functools.partial(mm_test.download)


meta = pysat.Meta()
meta['uts'] = {'units': 's',
               'long_name': 'Universal Time',
               'custom': False}
meta['Epoch'] = {'units': 'Milliseconds since 1970-1-1',
                 'Bin_Location': 0.5,
                 'notes': 'UTC time at middle of geophysical measurement.',
                 'desc': 'UTC seconds', }
meta['mlt'] = {'units': 'hours',
               'long_name': 'Magnetic Local Time',
               'label': 'MLT',
               'axis': 'MLT',
               'desc': 'Magnetic Local Time',
               'value_min': 0.0,
               'value_max': 24.0,
               'notes': ('Magnetic Local Time is the solar local time of the '
                         'field line at the location where the field crosses '
                         'the magnetic equator. In this case we just simulate '
                         '0-24 with a consistent orbital period and an offste '
                         'with SLT.'),
               'fill': np.nan,
               'scale': 'linear'}
meta['slt'] = {'units': 'hours',
               'long_name': 'Solar Local Time',
               'label': 'SLT',
               'axis': 'SLT',
               'desc': 'Solar Local Time',
               'value_min': 0.0,
               'value_max': 24.0,
               'notes': ('Solar Local Time is the local time (zenith angle of '
                         'sun) of the given location. Overhead noon, +/- 90 '
                         'is 6, 18 SLT .'),
               'fill': np.nan,
               'scale': 'linear'}
meta['orbit_num'] = {'units': '',
                     'long_name': 'Orbit Number',
                     'label': 'Orbit Number',
                     'axis': 'Orbit Number',
                     'desc': 'Orbit Number',
                     'value_min': 0.0,
                     'value_max': 25000.0,
                     'notes': ('Number of orbits since the start of the '
                               'mission. For this simulation we use the '
                               'number of 5820 second periods since the '
                               'start, 2008-01-01.'),
                     'fill': np.nan,
                     'scale': 'linear'}

meta['longitude'] = {'units': 'degrees', 'long_name': 'Longitude'}
meta['latitude'] = {'units': 'degrees', 'long_name': 'Latitude'}
meta['dummy1'] = {'units': '', 'long_name': 'dummy1'}
meta['dummy2'] = {'units': '', 'long_name': 'dummy2'}
meta['dummy3'] = {'units': '', 'long_name': 'dummy3'}
meta['dummy4'] = {'units': '', 'long_name': 'dummy4'}
meta['string_dummy'] = {'units': '', 'long_name': 'string_dummy'}
meta['unicode_dummy'] = {'units': '', 'long_name': 'unicode_dummy'}
meta['int8_dummy'] = {'units': '', 'long_name': 'int8_dummy'}
meta['int16_dummy'] = {'units': '', 'long_name': 'int16_dummy'}
meta['int32_dummy'] = {'units': '', 'long_name': 'int32_dummy'}
meta['int64_dummy'] = {'units': '', 'long_name': 'int64_dummy'}



"""
tests the pysat coords area
"""
import numpy as np
import warnings

import pandas as pds

from nose.tools import raises
import pysat
from pysat.utils import coords, time


class TestBasics():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.test_angles = np.array([340.0, 348.0, 358.9, 0.5, 5.0, 9.87])

        self.testInst = pysat.Instrument(platform='pysat',
                                         name='testing',
                                         clean_level='clean')
        # Add longitude to the test instrument
        ones = np.ones(shape=len(self.test_angles))
        tind = time.create_datetime_index(year=ones*2001,
                                          month=ones,
                                          uts=np.arange(0.0, len(ones), 1.0))

        self.testInst.data = \
            pds.DataFrame(np.array([tind, self.test_angles]).transpose(),
                          index=tind, columns=["time", "longitude"])

        self.deg_units = ["deg", "degree", "degrees", "rad", "radian",
                          "radians", "h", "hr", "hrs", "hours"]
        self.dist_units = ["m", "km", "cm"]
        self.vel_units = ["m/s", "cm/s", "km/s", 'm s$^{-1}$', 'cm s$^{-1}$',
                          'km s$^{-1}$', 'm s-1', 'cm s-1', 'km s-1']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.test_angles, self.testInst
        del self.deg_units, self.dist_units, self.vel_units

    #####################################
    # Cyclic data conversions

    def test_adjust_cyclic_data_default(self):
        """ Test adjust_cyclic_data with default range """

        test_in = np.radians(self.test_angles) - np.pi
        test_angles = coords.adjust_cyclic_data(test_in)

        assert test_angles.max() < 2.0 * np.pi
        assert test_angles.min() >= 0.0

    def test_adjust_cyclic_data_custom(self):
        """ Test adjust_cyclic_data with a custom range """

        test_angles = coords.adjust_cyclic_data(self.test_angles,
                                                high=180.0, low=-180.0)

        assert test_angles.max() < 180.0
        assert test_angles.min() >= -180.0

    #####################################
    # Update Longitude

    def test_update_longitude(self):
        """Test update_longitude """

        coords.update_longitude(self.testInst, lon_name="longitude")

        assert np.all(self.testInst.data['longitude'] < 180.0)
        assert np.all(self.testInst.data['longitude'] >= -180.0)

    @raises(ValueError)
    def test_bad_lon_name_update_longitude(self):
        """Test update_longitude with a bad longitude name"""

        coords.update_longitude(self.testInst, lon_name="not longitude")

    #########################
    # calc_solar_local_time

    def test_calc_solar_local_time(self):
        """Test calc_solar_local_time"""

        coords.calc_solar_local_time(self.testInst, lon_name="longitude",
                                     slt_name='slt')
        target = np.array([22.66666667, 23.20027778, 23.92722222,
                           0.03416667, 0.33444444, 0.65938889])

        assert (abs(self.testInst['slt'] - target)).max() < 1.0e-6

    def test_calc_solar_local_time_w_update_longitude(self):
        """Test calc_solar_local_time with update_longitude"""

        coords.calc_solar_local_time(self.testInst, lon_name="longitude",
                                     slt_name='slt')
        coords.update_longitude(self.testInst, lon_name="longitude")
        coords.calc_solar_local_time(self.testInst, lon_name="longitude",
                                     slt_name='slt2')

        assert (abs(self.testInst['slt']
                    - self.testInst['slt2'])).max() < 1.0e-6

    @raises(ValueError)
    def test_bad_lon_name_calc_solar_local_time(self):
        """Test calc_solar_local_time with a bad longitude name"""

        coords.calc_solar_local_time(self.testInst, lon_name="not longitude",
                                     slt_name='slt')

    def test_deprecation_warning_scale_units(self):
        """Test deprecation warning for this function"""

        import warnings

        warnings.simplefilter("always")
        scale1 = pysat.utils.scale_units("happy", "happy")
        with warnings.catch_warnings(record=True) as war:
            scale2 = coords.scale_units("happy", "happy")

        assert scale1 == scale2
        assert len(war) == 1
        assert war[0].category == DeprecationWarning

    ###################################
    # Geodetic / Geocentric conversions

    def test_geodetic_to_geocentric_single(self):
        """Test conversion from geodetic to geocentric coordinates"""

        lat, lon, rad = coords.geodetic_to_geocentric(45.0, lon_in=8.0)

        assert abs(lat - 44.807576784018046) < 1.0e-6
        assert abs(lon - 8.0) < 1.0e-6
        assert abs(rad - 6367.489543863465) < 1.0e-6

    def test_geocentric_to_geodetic_single(self):
        """Test conversion from geocentric to geodetic coordinates"""

        lat, lon, rad = coords.geodetic_to_geocentric(45.0, lon_in=8.0,
                                                      inverse=True)

        assert abs(lat - 45.192423215981954) < 1.0e-6
        assert abs(lon - 8.0) < 1.0e-6
        assert abs(rad - 6367.345908499981) < 1.0e-6

    def test_geodetic_to_geocentric_mult(self):
        """Test array conversion from geodetic to geocentric coordinates"""

        arr = np.ones(shape=(10,), dtype=float)
        lat, lon, rad = coords.geodetic_to_geocentric(45.0*arr, lon_in=8.0*arr)

        assert lat.shape == arr.shape
        assert lon.shape == arr.shape
        assert rad.shape == arr.shape
        assert abs(lat - 44.807576784018046).max() < 1.0e-6
        assert abs(lon - 8.0).max() < 1.0e-6
        assert abs(rad - 6367.489543863465).max() < 1.0e-6

    def test_geocentric_to_geodetic_mult(self):
        """Test array conversion from geocentric to geodetic coordinates"""

        arr = np.ones(shape=(10,), dtype=float)
        lat, lon, rad = coords.geodetic_to_geocentric(45.0*arr, lon_in=8.0*arr,
                                                      inverse=True)

        assert lat.shape == arr.shape
        assert lon.shape == arr.shape
        assert rad.shape == arr.shape
        assert abs(lat - 45.192423215981954).max() < 1.0e-6
        assert abs(lon - 8.0).max() < 1.0e-6
        assert abs(rad - 6367.345908499981).max() < 1.0e-6

    def test_geodetic_to_geocentric_inverse(self):
        """Tests the reversibility of geodetic to geocentric conversions"""

        lat1 = 37.5
        lon1 = 117.3
        lat2, lon2, rad_e = coords.geodetic_to_geocentric(lat1, lon_in=lon1,
                                                          inverse=False)
        lat3, lon3, rad_e = coords.geodetic_to_geocentric(lat2, lon_in=lon2,
                                                          inverse=True)
        assert abs(lon1-lon3) < 1.0e-6
        assert abs(lat1-lat3) < 1.0e-6

    ###############################################
    # Geodetic / Geocentric Horizontal conversions

    def test_geodetic_to_geocentric_horz_single(self):
        """Test conversion from geodetic to geocentric coordinates"""

        lat, lon, rad, az, el = \
            coords.geodetic_to_geocentric_horizontal(45.0, 8.0, 52.0, 63.0)

        assert abs(lat - 44.807576784018046) < 1.0e-6
        assert abs(lon - 8.0) < 1.0e-6
        assert abs(rad - 6367.489543863465) < 1.0e-6
        assert abs(az - 51.70376774257361) < 1.0e-6
        assert abs(el - 62.8811403841008) < 1.0e-6

    def test_geocentric_to_geodetic_horz_single(self):
        """Test conversion from geocentric to geodetic coordinates"""

        lat, lon, rad, az, el = \
            coords.geodetic_to_geocentric_horizontal(45.0, 8.0, 52.0, 63.0,
                                                     inverse=True)

        assert abs(lat - 45.192423215981954) < 1.0e-6
        assert abs(lon - 8.0) < 1.0e-6
        assert abs(rad - 6367.345908499981) < 1.0e-6
        assert abs(az - 52.29896101551479) < 1.0e-6
        assert abs(el - 63.118072033649916) < 1.0e-6

    def test_geodetic_to_geocentric_horz_mult(self):
        """Test array conversion from geodetic to geocentric coordinates"""

        arr = np.ones(shape=(10,), dtype=float)
        lat, lon, rad, az, el = \
            coords.geodetic_to_geocentric_horizontal(45.0*arr, 8.0*arr,
                                                     52.0*arr, 63.0*arr)

        assert lat.shape == arr.shape
        assert lon.shape == arr.shape
        assert rad.shape == arr.shape
        assert az.shape == arr.shape
        assert el.shape == arr.shape
        assert abs(lat - 44.807576784018046).max() < 1.0e-6
        assert abs(lon - 8.0).max() < 1.0e-6
        assert abs(rad - 6367.489543863465).max() < 1.0e-6
        assert abs(az - 51.70376774257361).max() < 1.0e-6
        assert abs(el - 62.8811403841008).max() < 1.0e-6

    def test_geocentric_to_geodetic_horz_mult(self):
        """Test array conversion from geocentric to geodetic coordinates"""

        arr = np.ones(shape=(10,), dtype=float)
        lat, lon, rad, az, el = \
            coords.geodetic_to_geocentric_horizontal(45.0*arr, 8.0*arr,
                                                     52.0*arr, 63.0*arr,
                                                     inverse=True)

        assert lat.shape == arr.shape
        assert lon.shape == arr.shape
        assert rad.shape == arr.shape
        assert az.shape == arr.shape
        assert el.shape == arr.shape
        assert abs(lat - 45.192423215981954).max() < 1.0e-6
        assert abs(lon - 8.0).max() < 1.0e-6
        assert abs(rad - 6367.345908499981).max() < 1.0e-6
        assert abs(az - 52.29896101551479).max() < 1.0e-6
        assert abs(el - 63.118072033649916).max() < 1.0e-6

    def test_geodetic_to_geocentric_horizontal_inverse(self):
        """Tests the reversibility of geodetic to geocentric horiz conversions

        Note:  inverse of az and el angles currently non-functional"""

        lat1 = -17.5
        lon1 = 187.3
        az1 = 52.0
        el1 = 63.0
        lat2, lon2, rad_e, az2, el2 = \
            coords.geodetic_to_geocentric_horizontal(lat1, lon1, az1, el1,
                                                     inverse=False)
        lat3, lon3, rad_e, az3, el3 = \
            coords.geodetic_to_geocentric_horizontal(lat2, lon2, az2, el2,
                                                     inverse=True)

        assert abs(lon1-lon3) < 1.0e-6
        assert abs(lat1-lat3) < 1.0e-6
        assert abs(az1-az3) < 1.0e-6
        assert abs(el1-el3) < 1.0e-6

    ####################################
    # Spherical / Cartesian conversions

    def test_spherical_to_cartesian_single(self):
        """Test conversion from spherical to cartesian coordinates"""

        x, y, z = coords.spherical_to_cartesian(45.0, 30.0, 1.0)

        assert abs(x - y) < 1.0e-6
        assert abs(z - 0.5) < 1.0e-6

    def test_cartesian_to_spherical_single(self):
        """Test conversion from cartesian to spherical coordinates"""

        x = 0.6123724356957946
        az, el, r = coords.spherical_to_cartesian(x, x, 0.5,
                                                  inverse=True)

        assert abs(az - 45.0) < 1.0e-6
        assert abs(el - 30.0) < 1.0e-6
        assert abs(r - 1.0) < 1.0e-6

    def test_spherical_to_cartesian_mult(self):
        """Test array conversion from spherical to cartesian coordinates"""

        arr = np.ones(shape=(10,), dtype=float)
        x, y, z = coords.spherical_to_cartesian(45.0*arr, 30.0*arr, arr)

        assert x.shape == arr.shape
        assert y.shape == arr.shape
        assert z.shape == arr.shape
        assert abs(x - y).max() < 1.0e-6
        assert abs(z - 0.5).max() < 1.0e-6

    def test_cartesian_to_spherical_mult(self):
        """Test array conversion from cartesian to spherical coordinates"""

        arr = np.ones(shape=(10,), dtype=float)
        x = 0.6123724356957946
        az, el, r = coords.spherical_to_cartesian(x*arr, x*arr, 0.5*arr,
                                                  inverse=True)

        assert az.shape == arr.shape
        assert el.shape == arr.shape
        assert r.shape == arr.shape
        assert abs(az - 45.0).max() < 1.0e-6
        assert abs(el - 30.0).max() < 1.0e-6
        assert abs(r - 1.0).max() < 1.0e-6

    def test_spherical_to_cartesian_inverse(self):
        """Tests the reversibility of spherical to cartesian conversions"""

        x1 = 3000.0
        y1 = 2000.0
        z1 = 2500.0
        az, el, r = coords.spherical_to_cartesian(x1, y1, z1,
                                                  inverse=True)
        x2, y2, z2 = coords.spherical_to_cartesian(az, el, r,
                                                   inverse=False)

        assert abs(x1-x2) < 1.0e-6
        assert abs(y1-y2) < 1.0e-6
        assert abs(z1-z2) < 1.0e-6

    ########################################
    # Global / Local Cartesian conversions

    def test_global_to_local_cartesian_single(self):
        """Test conversion from global to local cartesian coordinates"""

        x, y, z = coords.global_to_local_cartesian(7000.0, 8000.0, 9000.0,
                                                   37.5, 289.0, 6380.0)

        assert abs(x + 9223.175264852474) < 1.0e-6
        assert abs(y + 2239.835278362686) < 1.0e-6
        assert abs(z - 11323.126851088331) < 1.0e-6

    def test_local_cartesian_to_global_single(self):
        """Test conversion from local cartesian to global coordinates"""

        x, y, z = coords.global_to_local_cartesian(7000.0, 8000.0, 9000.0,
                                                   37.5, 289.0, 6380.0,
                                                   inverse=True)

        assert abs(x + 5709.804676635975) < 1.0e-6
        assert abs(y + 4918.397556010223) < 1.0e-6
        assert abs(z - 15709.577500484005) < 1.0e-6

    def test_global_to_local_cartesian_mult(self):
        """Test array conversion from global to local cartesian coordinates"""

        arr = np.ones(shape=(10,), dtype=float)
        x, y, z = coords.global_to_local_cartesian(7000.0*arr, 8000.0*arr,
                                                   9000.0*arr, 37.5*arr,
                                                   289.0*arr, 6380.0*arr)

        assert x.shape == arr.shape
        assert y.shape == arr.shape
        assert z.shape == arr.shape
        assert abs(x + 9223.175264852474).max() < 1.0e-6
        assert abs(y + 2239.835278362686).max() < 1.0e-6
        assert abs(z - 11323.126851088331).max() < 1.0e-6

    def test_local_cartesian_to_global_mult(self):
        """Test array conversion from local cartesian to global coordinates"""

        arr = np.ones(shape=(10,), dtype=float)
        x, y, z = coords.global_to_local_cartesian(7000.0*arr, 8000.0*arr,
                                                   9000.0*arr, 37.5*arr,
                                                   289.0*arr, 6380.0*arr,
                                                   inverse=True)

        assert x.shape == arr.shape
        assert y.shape == arr.shape
        assert z.shape == arr.shape
        assert abs(x + 5709.804676635975).max() < 1.0e-6
        assert abs(y + 4918.397556010223).max() < 1.0e-6
        assert abs(z - 15709.577500484005).max() < 1.0e-6

    def test_global_to_local_cartesian_inverse(self):
        """Tests the reversibility of the global to loc cartesian transform"""

        x1 = 7000.0
        y1 = 8000.0
        z1 = 9500.0
        lat = 37.5
        lon = 289.0
        rad = 6380.0
        x2, y2, z2 = coords.global_to_local_cartesian(x1, y1, z1,
                                                      lat, lon, rad,
                                                      inverse=False)
        x3, y3, z3 = coords.global_to_local_cartesian(x2, y2, z2,
                                                      lat, lon, rad,
                                                      inverse=True)
        assert abs(x1-x3) < 1.0e-6
        assert abs(y1-y3) < 1.0e-6
        assert abs(z1-z3) < 1.0e-6

    ########################################
    # Local Horizontal / Global conversions

    def test_local_horizontal_to_global_geo_geodetic(self):
        """Tests the conversion of the local horizontal to global geo"""

        az = 30.0
        el = 45.0
        dist = 1000.0
        lat0 = 45.0
        lon0 = 0.0
        alt0 = 400.0

        lat, lon, rad = \
            coords.local_horizontal_to_global_geo(az, el, dist,
                                                  lat0, lon0, alt0)

        assert abs(lat - 50.419037572472625) < 1.0e-6
        assert abs(lon + 7.694008395350697) < 1.0e-6
        assert abs(rad - 7172.15486518744) < 1.0e-6

    def test_local_horizontal_to_global_geo(self):
        """Tests the conversion of the local horizontal to global geo"""

        az = 30.0
        el = 45.0
        dist = 1000.0
        lat0 = 45.0
        lon0 = 0.0
        alt0 = 400.0

        lat, lon, rad = \
            coords.local_horizontal_to_global_geo(az, el, dist,
                                                  lat0, lon0, alt0,
                                                  geodetic=False)

        assert abs(lat - 50.414315865044202) < 1.0e-6
        assert abs(lon + 7.6855551809119502) < 1.0e-6
        assert abs(rad - 7185.6983665760772) < 1.0e-6


class TestDeprecation():

    def setup(self):
        """Runs before every method to create a clean testing setup"""
        warnings.simplefilter("always")

    def teardown(self):
        """Runs after every method to clean up previous testing"""

    def test_deprecation_warning_geodetic_to_geocentric(self):
        """Test if geodetic_to_geocentric in coords is deprecated"""

        with warnings.catch_warnings(record=True) as war:
            coords.geodetic_to_geocentric(45.0, lon_in=8.0)

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning

    def test_deprecation_warning_geodetic_to_geocentric_horz(self):
        """Test if geodetic_to_geocentric_horizontal in coords is deprecated"""

        with warnings.catch_warnings(record=True) as war:
            coords.geodetic_to_geocentric_horizontal(45.0, 8.0, 52.0, 63.0)

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning

    def test_deprecation_warning_spherical_to_cartesian(self):
        """Test if spherical_to_cartesian in coords is deprecated"""

        with warnings.catch_warnings(record=True) as war:
            coords.spherical_to_cartesian(45.0, 30.0, 1.0)

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning

    def test_deprecation_warning_global_to_local_cartesian(self):
        """Test if global_to_local_cartesian in coords is deprecated"""

        with warnings.catch_warnings(record=True) as war:
            coords.global_to_local_cartesian(7000.0, 8000.0, 9000.0,
                                             37.5, 289.0, 6380.0)

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning

    def test_deprecation_warning_local_horizontal_to_global_geo(self):
        """Test if local_horizontal_to_global_geo in coords is deprecated"""

        with warnings.catch_warnings(record=True) as war:
            coords.local_horizontal_to_global_geo(30.0, 45.0, 1000.0,
                                                  45.0, 0.0, 400.0)
        assert len(war) >= 1
        assert war[0].category == DeprecationWarning



# -*- coding: utf-8 -*-
"""Supports the Retarding Potential Analyzer (RPA) instrument on
Dynamics Explorer 2 (DE2).

From CDAWeb:

The Retarding Potential Analyzer (RPA) measured the bulk ion velocity in the
direction of the spacecraft motion, the constituent ion concentrations, and the
ion temperature along the satellite path. These parameters were derived from a
least squares fit to the ion number flux vs energy curve obtained by sweeping
or stepping the voltage applied to the internal retarding grids of the RPA. In
addition, a separate wide aperture sensor, a duct sensor, was flown to measure
the spectral characteristics of iregularities in the total ion concentration.
The measured parameters obtained from this investigation were important to the
understanding of mechanisms that influence the plasma; i.e., to understand the
coupling between the solar wind and the earth's atmosphere. The measurements
were made with a multigridded planar retarding potential analyzer very similar
in concept and geometry to the instruments carried on the AE satellites. The
retarding potential was variable in the range from approximately +32 to 0 V.
The details of this voltage trace, and whether it was continuous or stepped,
depended on the operating mode of the instrument. Specific parameters deduced
from these measurements were ion temperature; vehicle potential; ram component
of the ion drift velocity; the ion and electron concentration irregularity
spectrum; and the concentration of H+, He+, O+, and Fe+, and of molecular ions
near perigee.

It includes the DUCT portion of the high resolutiondata from the Dynamics
Explorer 2 (DE-2) Retarding Potential Analyzer (RPA) for the whole DE-2 mission
time period in ASCII format. This version was generated at NSSDC from the
PI-provided binary data (SPIO-00232). The DUCT files include RPA measurements
ofthe total ion concentration every 64 times per second. Due to a failure in
the instrument memory system RPA data are not available from 81317 06:26:40 UT
to 82057 13:16:00 UT. This data set is based on the revised version of the RPA
files that was submitted by the PI team in June of 1995. The revised RPA data
include a correction to the spacecraft potential.

References
----------
W. B. Hanson, R. A. Heelis, R. A. Power, C. R. Lippincott, D. R. Zuccaro,
B. J. Holt, L. H. Harmon, and S. Sanatani, “The retarding potential analyzer
for dynamics explorer-B,” Space Sci. Instrum. 5, 503–510 (1981).

Properties
----------
platform
    'de2'
name
    'rpa'
sat_id
    ''
tag
    None Supported

Authors
-------
J. Klenzing

"""

from __future__ import print_function
from __future__ import absolute_import

import functools

import pysat
from pysat.instruments.methods import nasa_cdaweb as cdw
from pysat.instruments.methods import general as mm_gen

platform = 'de2'
name = 'rpa'

tags = {'': '2 sec cadence RPA data'}  # this is the default
sat_ids = {'': ['']}
_test_dates = {'': {'': pysat.datetime(1983, 1, 1)}}

fname = 'de2_ion2s_rpa_{year:04d}{month:02d}{day:02d}_v01.cdf'
supported_tags = {'': {'': fname}}

# use the CDAWeb methods list files routine
list_files = functools.partial(mm_gen.list_files,
                               supported_tags=supported_tags)

# use the default CDAWeb method
load = cdw.load

# support download routine
basic_tag = {'dir': '/pub/data/de/de2/plasma_rpa/ion2s_cdaweb',
             'remote_fname': '{year:4d}/' + fname,
             'local_fname': fname}
supported_tags = {'': {'': basic_tag}}
download = functools.partial(cdw.download, supported_tags)

# support listing files currently on CDAWeb
list_remote_files = functools.partial(cdw.list_remote_files,
                                      supported_tags=supported_tags)


def clean(inst):
    """Routine to return PLATFORM/NAME data cleaned to the specified level

    Cleaning level is specified in inst.clean_level and pysat
    will accept user input for several strings. The clean_level is
    specified at instantiation of the Instrument object.

    'clean' All parameters should be good, suitable for statistical and
            case studies
    'dusty' All paramers should generally be good though same may
            not be great
    'dirty' There are data areas that have issues, data should be used
            with caution
    'none'  No cleaning applied, routine not called in this case.


    Parameters
    -----------
    inst : pysat.Instrument
        Instrument class object, whose attribute clean_level is used to return
        the desired level of data selectivity.

    """

    return



#!/usr/bin/env python
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.1199703
# ----------------------------------------------------------------------------

import copy
import datetime as dt
import functools
import numpy as np
import pandas as pds
import xarray as xr
import weakref

from pysat import logger


class Orbits(object):
    """Determines orbits on the fly and provides orbital data in .data.

    Parameters
    ----------
    inst : pysat.Instrument
        Instrument object for which the orbits will be determined
    index : str or NoneType
        Name of the data series to use for determing orbit breaks (default=None)
    kind : str
        Kind of orbit, which specifies how orbital breaks are determined.
        Expects one of: 'local time', 'longitude', 'polar', or 'orbit'
        - local time: negative gradients in lt or breaks in inst.data.index
        - longitude: negative gradients or breaks in inst.data.index
        - polar: zero crossings in latitude or breaks in inst.data.index
        - orbit: uses unique values of orbit number
        (default='local time')
    period : np.timedelta64 or NoneType
        length of time for orbital period, used to gauge when a break
        in the datetime index (inst.data.index) is large enough to
        consider it a new orbit

    Attributes
    ----------
    inst : pysat.Instrument
        Instrument object for which the orbits will be determined
    kind : str
        Kind of orbit, which specifies how orbital breaks are determined.
        Expects one of: 'local time', 'longitude', 'polar', or 'orbit'
        - local time : negative gradients in lt or breaks in inst.data.index
        - longitude : negative gradients or breaks in inst.data.index
        - polar : zero crossings in latitude or breaks in inst.data.index
        - orbit : uses unique values of orbit number
        (default='local time')
    orbit_period : pds.Timedelta
        Pandas Timedelta that specifies the orbit period.  Used instead of
        dt.timedelta to enable np.timedelta64 input.
        (default=97 min)
    num : int
        Number of orbits in loaded data
    orbit_index : int
        Index of currently loaded orbit, zero indexed

    Note
    ----
    Determines the locations of orbit breaks in the loaded data in inst.data
    and provides iteration tools and convenient orbit selection via
    inst.orbit[orbit num].

    This class should not be called directly by the user, it uses the interface
    provided by inst.orbits where inst = pysat.Instrument()

    Examples
    --------
    ::

        # Use orbit_info Instrument keyword to pass all Orbit kwargs
        orbit_info = {'index': 'longitude', 'kind': 'longitude'}
        vefi = pysat.Instrument(platform='cnofs', name='vefi', tag='dc_b',
                                clean_level=None, orbit_info=orbit_info)

        # Set the instrument bounds
        start = dt.datetime(2009, 1, 1)
        stop = dt.datetime(2009, 1, 10)
        vefi.load(date=start)
        vefi.bounds(start, stop)

        # Iterate over orbits
        for vefi in vefi.orbits:
            print('Next available orbit ', vefi['dB_mer'])

        # Load fifth orbit of first day
        vefi.load(date=start)
        vefi.orbits[5]

        # Less convenient load
        vefi.orbits.load(5)

        # Manually iterate forwards in the orbit
        vefi.orbits.next()

        # Manually iterate backwards in the orbit
        vefi.orbits.prev()

    """

    # -----------------------------------------------------------------------
    # Define the magic methods

    def __init__(self, inst, index=None, kind='local time', period=None):

        # Set the class attributes
        self.inst = weakref.proxy(inst)
        self.kind = kind.lower()

        if period is None:
            period = pds.Timedelta(np.timedelta64(97, 'm'))
        self.orbit_period = pds.Timedelta(period)

        orbit_breaks = None
        if self.kind in ['local time', 'lt']:
            orbit_breaks = 24.0
        elif self.kind in ['longitude', 'long', 'lon']:
            orbit_breaks = 360.0

        if orbit_breaks is None:
            if self.kind == 'polar':
                self._det_breaks = self._polar_breaks
            elif self.kind == 'orbit':
                self._det_breaks = self._orbit_number_breaks
            else:
                raise ValueError('Unknown kind of orbit requested.')
        else:
            self._det_breaks = functools.partial(
                self._equa_breaks, orbit_index_period=orbit_breaks)

        self._orbit_breaks = []
        self.num = 0
        self._current = 0
        self.orbit_index = index

    def __repr__(self):
        """ Print the basic Orbits properties"""
        out_str = "".join(["pysat.Orbits(inst=", self.inst.__repr__(),
                           ", index=", self.orbit_index.__repr__(),
                           ", kind=", self.kind.__repr__(), ", period=",
                           self.orbit_period.__repr__(), ")"])
        return out_str

    def __str__(self):
        """ Descriptively print the basic Obits properties"""
        output_str = 'Orbit Settings\n'
        output_str += '--------------\n'
        output_str += 'Orbit Kind: {:s}\n'.format(self.kind.__repr__())
        output_str += 'Orbit Index: {:s}\n'.format(self.orbit_index.__repr__())
        output_str += 'Orbit Period: {:s}\n'.format(
            self.orbit_period.__repr__())
        output_str += 'Number of Orbits: {:d}\n'.format(self.num)
        output_str += 'Loaded Orbit Number: {:s}\n'.format(
            self._current.__repr__())

        return output_str

    def __eq__(self, other):
        """Perform equality check

        Parameters
        ----------
        other : any
            Other object to compare for equality

        Returns
        -------
        bool
            True if objects are identical, False if they are not

        """

        # Check if other is the same class (Orbits). Exit early if not.
        if not isinstance(other, self.__class__):
            return False

        # If the type is the same then check everything that is attached to
        # the Orbits object. Includes attributes, methods, variables, etc.
        checks = []
        key_check = []
        for key in self.__dict__.keys():
            if key in other.__dict__.keys():
                if key not in ['_full_day_data', 'inst', '_det_breaks']:
                    # Standard equality comparison
                    test = np.all(self.__dict__[key] == other.__dict__[key])
                    checks.append(test)
                    key_check.append(key)

                elif key in ['_full_day_data']:
                    # Compare data
                    if isinstance(self.__dict__[key], pds.DataFrame):
                        try:
                            # Comparisons can error simply for having
                            # different DataFrames
                            check = np.all(self.__dict__[key]
                                           == other.__dict__[key])
                        except ValueError:
                            # If there is an error they aren't the same
                            return False

                        checks.append(check)
                        key_check.append(key)

                    else:
                        # xarray comparison
                        test = xr.Dataset.equals(self.__dict__[key],
                                                 other.__dict__[key])
                        checks.append(test)
                        key_check.append(key)

                elif key == '_det_breaks':
                    # Equality of partial functions does not work well.
                    # Using a string comparison instead. This can also break
                    # if one of the objects is missing some attributes.
                    try:
                        check = str(self._det_breaks) == str(other._det_breaks)
                    except AttributeError:
                        # One object is missing a required attribute
                        return False

                    checks.append(check)
                    key_check.append(key)

            else:
                checks.append(False)
                key_check.append(key)
                return False

        # Confirm that Orbits object `other` doesn't have extra terms
        for key in other.__dict__.keys():
            if key not in self.__dict__.keys():
                return False

        test_data = np.all(checks)

        return test_data

    def __getitem__(self, orbit_key):
        """Enable convenience notation for loading orbit into parent object.

        Parameters
        ----------
        orbit_key : int or None
            Orbit number to get

        Examples
        --------
        ::

            inst.load(date=date)
            inst.orbits[4]
            print('Orbit data ', inst.data)

        Note
        ----
        A day of data must already be loaded.

        """
        if orbit_key < 0:
            self.load(orbit_key)  # Loading for reverse indices
        else:
            self.load(orbit_key + 1)  # Loading for forward indices

    def __iter__(self):
        """Support iteration by orbit.

        Examples
        --------
        ::

            for inst in inst.orbits:
                print('next available orbit ', inst.data)

        Note
        ----
        For each iteration the next available orbit is loaded into
        inst.data.

        Limits of iteration set by setting inst.bounds.

        """
        # Load up the first increment of data
        while self.inst.empty:
            self.inst.next()

        while True:
            try:
                self.next()
                yield self.inst
            except StopIteration:
                return

    # -----------------------------------------------------------------------
    # Define the hidden methods

    def _report_current_orbit(self):
        """ Report the current orbit to log at the info level
        """

        # Index appears as zero-indexed, though it is one-indexed
        logger.info('Loaded Orbit: {:d}'.format(self._current - 1))
        return

    def _reset(self):
        """Create null arrays for storing orbit info
        """
        self._orbit_breaks = []
        self.num = 0
        self._current = 0

    def _calc_orbits(self):
        """Prepares data structure for breaking data into orbits

        Raises
        ------
        ValueError
            If the the Instrument bounds are set to load overlapping data sets

        """
        # Check there isn't an overlapping data set from iteration bounds
        estr = ' '.join(('Orbit iteration is not currently supported',
                         'when the pysat.Instrument bounds are',
                         'configured for loading overlapping',
                         'data. Please set the Instrument bounds width',
                         'to be less than or equal to the bounds step ',
                         'increment. See `pysat.Instrument.bounds` for more.'))

        if self.inst._iter_type == 'file':
            if self.inst._iter_step < self.inst._iter_width:
                raise ValueError(estr)
        else:
            # Iterating by date.  We need to check step (frequency string)
            # against width (timedelta)
            step = pds.tseries.frequencies.to_offset(self.inst._iter_step)
            step = dt.timedelta(seconds=step.delta.total_seconds())
            root = dt.datetime(2001, 1, 1)
            if root + step < root + self.inst._iter_width:
                raise ValueError(estr)

        # If the breaks between orbit have not been defined, define them here.
        # Also store the data so that grabbing different orbits does not
        # require reloads of whole dataset
        if len(self._orbit_breaks) == 0:
            # Eetermine orbit breaks
            self._det_breaks()

            # Store a copy of data
            self._full_day_data = self.inst.data.copy()

            # Set current orbit counter to zero (default)
            self._current = 0

    def _equa_breaks(self, orbit_index_period=24.0):
        """Determine where breaks in an equatorial satellite orbit occur.

        Looks for negative gradients in local time (or longitude) as well as
        breaks in UT.

        Parameters
        ----------
        orbit_index_period : float
            The change in value of supplied index parameter for a single orbit
            (default=24.0)

        Raises
        ------
        ValueError
            If the orbit_index attribute is not set to an appropriate value

        """

        if self.orbit_index is None:
            raise ValueError(' '.join(('Orbit properties must be defined at ',
                                       'pysat.Instrument object instantiation.',
                                       'See Instrument docs.')))
        else:
            try:
                self.inst[self.orbit_index]
            except KeyError as err:
                raise ValueError(''.join((str(err), '\n',
                                          'Provided orbit index does not ',
                                          'exist in loaded data')))

        # Get the difference in orbit index around the orbit
        lt_diff = self.inst[self.orbit_index]
        if not self.inst.pandas_format:
            lt_diff = lt_diff.to_pandas()
        lt_diff = lt_diff.diff()

        # Get the typical (median) difference
        typical_lt_diff = np.nanmedian(lt_diff)
        logger.info(''.join(('typical lt diff ', str(typical_lt_diff))))

        # Get the Universal Time difference between data values. Assumes that
        # the time index is in UT.
        ut_vals = pds.Series(self.inst.index)
        ut_diff = ut_vals.diff()

        # Get the locations where the orbit index derivative is less than 0,
        # then do some sanity checks on these locations
        ind, = np.where((lt_diff < -0.2 * typical_lt_diff))
        if len(ind) > 0:
            ind = np.hstack((ind, np.array([len(self.inst[self.orbit_index])])))

            # Look at distance between breaks
            dist = ind[1:] - ind[0:-1]

            # Only keep orbit breaks with a distance greater than 1.  This check
            # is done to ensure robustness
            if len(ind) > 1:
                if min(dist) == 1:
                    logger.info(' '.join(('There are orbit breaks right next',
                                          'to each other')))
                ind = ind[:-1][dist > 1]

            # Check for large positive gradients around the break that would
            # suggest not a true orbit break, but rather bad orbit_index values
            new_ind = []
            for idx in ind:
                tidx, = np.where(lt_diff[(idx - 5):(idx + 6)]
                                 > 10 * typical_lt_diff)

                if len(tidx) != 0:
                    # There are large changes, this suggests a false alarm.
                    # Iterate over samples and check
                    for sub_tidx in tidx:
                        # Look at time change vs local time change
                        if(ut_diff[idx - 5:idx + 6].iloc[sub_tidx]
                           < lt_diff[idx - 5:idx + 6].iloc[sub_tidx]
                           / orbit_index_period * self.orbit_period):
                            # The change in UT is small compared to the change
                            # in the orbit index this is flagged as a false
                            # alarm, or dropped from consideration
                            logger.info(''.join(('Dropping found break ',
                                                 'as false positive.')))
                            pass
                        else:
                            # The change in UT is significant, keep orbit break
                            new_ind.append(idx)
                            break
                else:
                    # There are no large positive gradients, current orbit
                    # break passes the first test
                    new_ind.append(idx)

            # Replace all breaks with those that are 'good'
            ind = np.array(new_ind)

        # Now, assemble some orbit breaks that are not triggered by changes in
        # the orbit index
        #
        # Check if there is a UT break that is larger than orbital period, AKA
        # a time gap
        ut_change_vs_period = (ut_diff > self.orbit_period)

        # Characterize ut change using orbital period
        norm_ut = ut_diff / self.orbit_period

        # Now, look for breaks because the length of time between samples is
        # too large, thus there is no break in slt/mlt/etc, lt_diff is small
        # but UT change is big
        norm_ut_vs_norm_lt = norm_ut.gt(np.abs(lt_diff.values
                                               / orbit_index_period))

        # Indices when one or other flag is true
        ut_ind, = np.where(ut_change_vs_period
                           | (norm_ut_vs_norm_lt & (norm_ut > 0.95)))

        # Combine these UT determined orbit breaks with the orbit index orbit
        # breaks
        if len(ut_ind) > 0:
            ind = np.hstack((ind, ut_ind))
            ind = np.sort(ind)
            ind = np.unique(ind)
            logger.info('Time Gap at locations: {:}'.format(ut_ind))

        # Now that most problems in orbits should have been caught, look at
        # the time difference between orbits (not individual orbits)
        orbit_ut_diff = ut_vals[ind].diff()
        if not self.inst.pandas_format:
            orbit_lt_diff = self.inst[self.orbit_index].to_pandas()[ind].diff()
        else:
            orbit_lt_diff = self.inst[self.orbit_index][ind].diff()

        # Look for time gaps between partial orbits. The full orbital time
        # period is not required between end of one orbit and beginning of next
        # if first orbit is partial.  Also provides another general test of the
        # orbital breaks determined.
        idx, = np.where((orbit_ut_diff / self.orbit_period
                         - orbit_lt_diff.values / orbit_index_period) > 0.97)

        # Pull out breaks that pass the test, need to make sure the first one
        # is always included it gets dropped via the nature of diff
        if len(idx) > 0:
            if idx[0] != 0:
                idx = np.hstack((0, idx))
        else:
            idx = np.array([0])

        # Only keep the good indices
        if len(ind) > 0:
            ind = ind[idx]

            # Create an orbit break index, ensure first element is always 0
            if ind[0] != 0:
                ind = np.hstack((np.array([0]), ind))
        else:
            ind = np.array([0])

        # Set the index of orbit breaks and the number of orbits
        self._orbit_breaks = ind
        self.num = len(ind)

    def _polar_breaks(self):
        """Determine where breaks in a polar orbiting satellite orbit occur.

        Raises
        ------
        ValueError
            If the orbit_index attribute is not set to an appropriate value

        Note
        ----
        Looks for sign changes in latitude (magnetic or geographic) as well as
        breaks in UT.

        """

        if self.orbit_index is None:
            raise ValueError(' '.join(('Orbit properties must be defined at',
                                       'pysat.Instrument object instantiation.',
                                       'See Instrument docs.')))
        else:
            try:
                self.inst[self.orbit_index]
            except KeyError as err:
                raise ValueError(''.join((str(err), '\n',
                                          'Provided orbit index does not ',
                                          'appear to exist in loaded data')))

        # Determine where orbit index goes from positive to negative
        pos = (self.inst[self.orbit_index] >= 0)
        npos = np.logical_not(pos)
        change = ((pos.values[:-1] & npos.values[1:])
                  | (npos.values[:-1] & pos.values[1:]))

        ind, = np.where(change)
        ind += 1

        ut_diff = pds.Series(self.inst.index).diff()
        ut_ind, = np.where(ut_diff / self.orbit_period > 0.95)

        if len(ut_ind) > 0:
            ind = np.unique(np.sort(np.hstack((ind, ut_ind))))

        # Create an orbit break index, ensure first element is always 0
        if len(ind) > 0:
            if ind[0] != 0:
                ind = np.hstack((np.array([0]), ind))
        else:
            ind = np.array([0])

        # Set the index of orbit breaks and the number of orbits
        self._orbit_breaks = ind
        self.num = len(ind)

    def _orbit_number_breaks(self):
        """Find orbital breaks in a dataset with orbit numbers occur.

        Raises
        ------
        ValueError
            If the orbit_index attribute is not set to an appropriate value

        Note
        ----
        Looks for changes in unique values.

        """

        if self.orbit_index is None:
            raise ValueError(' '.join(('Orbit properties must be defined at ',
                                       'pysat.Instrument object instantiation.',
                                       'See Instrument docs.')))
        else:
            try:
                self.inst[self.orbit_index]
            except KeyError as err:
                raise ValueError(''.join((str(err), '\n',
                                          'Provided orbit index does not ',
                                          'appear to exist in loaded data')))

        # Determine where the orbit index changes from one value to the next
        uniq_vals = np.unique(self.inst[self.orbit_index].values)
        orbit_index = []
        for val in uniq_vals:
            idx, = np.where(val == self.inst[self.orbit_index].values)
            orbit_index.append(idx[0])

        # Create orbit break index, ensure first element is always 0
        if len(orbit_index) > 0:
            if orbit_index[0] != 0:
                ind = np.hstack((np.array([0]), orbit_index))
            else:
                ind = orbit_index
        else:
            ind = np.array([0])

        # Set the index of orbit breaks and the number of orbits
        self._orbit_breaks = ind
        self.num = len(ind)

    def _get_basic_orbit(self, orbit_num):
        """Load a particular orbit into .data for loaded day.

        Parameters
        ----------
        orbit_num : int
            orbit number, 1 indexed, negative indexes allowed, -1 last orbit

        Note
        ----
        A day of data must be loaded before this routine functions properly.
        If the last orbit of the day is requested, it will NOT automatically be
        padded with data from the next day.

        """
        # Ensure data exists
        if not self.inst.empty:
            # Ensure proper orbit metadata present
            self._calc_orbits()

            # Set up data access for both pandas and xarray
            self.inst.data = self._full_day_data

            # Pull out the requested orbit
            if orbit_num == -1:
                # Load last orbit data into data
                orb_break = self._orbit_breaks[self.num + orbit_num]
                self.inst.data = self.inst[orb_break:]
                self._current = self.num + orbit_num + 1
            elif orbit_num < 0 and orbit_num >= -self.num:
                # Load backwards index orbit data into data
                self.inst.data = self.inst[
                    self._orbit_breaks[self.num + orbit_num]:
                    self._orbit_breaks[self.num + orbit_num + 1]]
                self._current = self.num + orbit_num + 1
            elif orbit_num < self.num and orbit_num != 0:
                # Load forward indexed orbit data into data
                self.inst.data = self.inst[self._orbit_breaks[orbit_num - 1]:
                                           self._orbit_breaks[orbit_num]]
                self._current = orbit_num
            elif orbit_num == self.num:
                self.inst.data = self.inst[self._orbit_breaks[orbit_num - 1]:]
                self._current = orbit_num
            elif orbit_num == 0:
                raise ValueError(' '.join(('Orbits internally indexed by',
                                           '1, 0 not allowed')))
            else:
                # Gone too far
                self.inst.data = []
                raise ValueError(' '.join(('Requested an orbit past total',
                                           'orbits for day')))
        return

    # -----------------------------------------------------------------------
    # Define the public methods and properties

    def copy(self):
        """Provide a deep copy of object

        Returns
        -------
        Orbits class instance
            Copy of self

        """
        # pysat.Instrument has a link to orbits, so copying the referenced
        # self.inst would lead to infinite recursion.
        inst = self.inst
        self.inst = None

        # Copy everything else
        orbits_copy = copy.deepcopy(self)

        # Both this object and the copy refer back to the same pysat.Instrument
        orbits_copy.inst = inst
        self.inst = inst

        return orbits_copy

    @property
    def current(self):
        """Current orbit number.

        Returns
        -------
        int or NoneType
            None if no orbit data. Otherwise, returns orbit number, beginning
            with zero. The first and last orbit of a day is somewhat ambiguous.
            The first orbit for day n is generally also the last orbit
            on day n - 1. When iterating forward, the orbit will be labeled
            as first (0). When iterating backward, orbit labeled as the last.

        """

        if self._current > 0:
            return self._current - 1
        else:
            return None

    def load(self, orbit_num):
        """Load a particular orbit into .data for loaded day.

        Parameters
        ----------
        orbit_num : int
            orbit number, 1 indexed (1-length or -1 to -length) with sign
            denoting forward or backward indexing

        Raises
        ------
        ValueError
            If index requested lies beyond the number of orbits

        Note
        ----
        A day of data must be loaded before this routine functions properly.
        If the last orbit of the day is requested, it will automatically be
        padded with data from the next day. The orbit counter will be
        reset to 1.

        """
        # Ensure data exits
        if not self.inst.empty:
            # Set up orbit metadata
            self._calc_orbits()

            # Pull out the requested orbit
            if orbit_num < 0:
                # Negative indexing consistent with numpy, -1 last,
                # -2 second to last, etc.
                orbit_num = self.num + 1 + orbit_num

            if orbit_num == self.num:
                # We get here if user asks for last orbit. This cal is first to
                # trap case where there is only one orbit (self.num=1), which
                # needs to be treated differently than a orbit=1 call
                if self.num != 1:
                    # More than one orbit, go back one (simple call) and
                    # then forward doing full logic for breaks across day
                    self._get_basic_orbit(self.num - 1)
                    self.next()
                else:
                    # At most one complete orbit in the file, check if we are
                    # close to beginning or end of day
                    date = self.inst.date
                    delta_start = self.inst.index[-1] - date
                    delta_end = (date + dt.timedelta(days=1)
                                 - self.inst.index[0])

                    if delta_start <= self.orbit_period * 1.05:
                        # We are near the beginning. Load the previous file,
                        # then go forward one orbit
                        self.inst.prev()
                        self.next()
                        if self.inst.index[-1] < date + delta_start:
                            # We could go back a day, iterate over orbit, as
                            # above, and the data we have is the wrong day.
                            # In this case, move forward again.  This happens
                            # when previous day doesn't have data near end of
                            # the day
                            self.next()

                    elif delta_end <= self.orbit_period * 1.05:
                        # Near end; load next file, then go back one orbit
                        self.inst.next()
                        self.prev()
                        if self.inst.index[0] > (date - delta_end
                                                 + dt.timedelta(days=1)):
                            # We could go forward a day, iterate over orbit
                            # as above, and the data we have is the wrong day.
                            # In this case, move back again. This happens when
                            # next day doesn't have data near beginning of the
                            # day
                            self.prev()
                    else:
                        # Not near beginning or end, just get the last orbit
                        # available (only one)
                        self._get_basic_orbit(-1)
            elif orbit_num == 1:
                # User asked for first orbit
                try:
                    # Orbit could start file previous; check for this condition
                    # and store the real date user wants
                    true_date = self.inst.date

                    # Go back a day
                    self.inst.prev()

                    # If and else added because of Instruments that have large
                    # gaps (e.g., C/NOFS).  In this case, prev can return
                    # empty data
                    if not self.inst.empty:
                        # Get last orbit if there is data. This will deal with
                        # orbits across file cleanly
                        self.load(-1)
                    else:
                        # No data, no previous data to account for. Move back
                        # to original data, do simple load of first orbit
                        self.inst.next()
                        self._get_basic_orbit(1)

                    # Check that this orbit should end on the current day
                    delta = true_date - self.inst.index[0]
                    if delta >= self.orbit_period:
                        # The orbit loaded isn't close enough to date to be the
                        # first orbit of the day, move forward
                        self.next()

                except StopIteration:
                    # Check if the first orbit is also the last orbit
                    self._get_basic_orbit(1)
                    self._report_current_orbit()

            elif orbit_num < self.num:
                # Load basic orbit data into data
                self._get_basic_orbit(orbit_num)
                self._report_current_orbit()

            else:
                # Gone too far
                self.inst.data = self.inst._null_data
                raise ValueError(' '.join(('Requested an orbit past total',
                                           'orbits for day')))
        else:
            logger.info(' '.join(('No data loaded in instrument object to',
                                  'determine orbits.')))

    def next(self):
        """Load the next orbit into associated Instrument.data object

        Note
        ----
        Forms complete orbits across day boundaries. If no data loaded
        then the first orbit from the first date of data is returned.

        """

        # Check if data exists
        if not self.inst.empty:
            # Set up orbit metadata
            self._calc_orbits()

            # If current orbit near the last, must be careful
            if self._current == (self.num - 1):
                # Load last orbit data
                self._get_basic_orbit(-1)

                # End of orbit may occur on the next day
                load_next = True
                if self.inst._iter_type == 'date':
                    delta = (self.inst.date - self.inst.index[-1]
                             + pds.Timedelta('1 day'))
                    if delta >= self.orbit_period:
                        # Don't need to load the next day because this orbit
                        # ends more than a orbital period from the next date
                        load_next = False

                if load_next:
                    # The end of the user's desired orbit occurs tomorrow, need
                    # to form a complete orbit save this current orbit, load
                    # the next day, combine data, select the correct orbit
                    temp_orbit_data = self.inst.copy()
                    try:
                        # Loading next day/file clears orbit breaks info
                        self.inst.next()
                        if not self.inst.empty:
                            # Combine this next day's data with previous last
                            # orbit, grab the first one
                            final_val = self.inst.index[0] - dt.timedelta(
                                microseconds=1)
                            self.inst.concat_data(temp_orbit_data[:final_val],
                                                  prepend=True)
                            self._get_basic_orbit(1)
                        else:
                            # No data, go back a day and grab the last orbit.
                            # This is as complete as this orbit can be.
                            self.inst.prev()
                            self._get_basic_orbit(-1)
                    except StopIteration:
                        pass
                    del temp_orbit_data

                self._report_current_orbit()

            elif self._current == (self.num):
                # At the last orbit, need to be careful about getting the next
                # orbit save this current orbit and load the next day
                # temp_orbit_data = self.inst.data.copy()
                temp_orbit_data = self.inst.copy()

                # Load next day, which clears orbit breaks info
                self.inst.next()

                # Combine this next day orbit with previous last orbit to
                # ensure things are correct
                if not self.inst.empty:
                    pad_next = True

                    # Check if data padding is really needed, only works when
                    # loading by date
                    if self.inst._iter_type == 'date':
                        delta = self.inst.date - temp_orbit_data.index[-1]
                        if delta >= self.orbit_period:
                            # The end of the previous orbit is more than an
                            # orbit away from today we don't have to worry
                            # about it
                            pad_next = False
                    if pad_next:
                        # The orbit went across day break, stick old orbit onto
                        # new data and grab second orbit (first is old)
                        self.inst.concat_data(
                            temp_orbit_data[:self.inst.index[0]
                                            - dt.timedelta(microseconds=1)],
                            prepend=True)

                        # Select second orbit of combined data
                        self._get_basic_orbit(2)
                    else:
                        # Padding from the previous orbit wasn't needed, can
                        # just grab the first orbit of loaded data
                        self._get_basic_orbit(1)
                        if self.inst._iter_type == 'date':
                            delta = (self.inst.date + dt.timedelta(days=1)
                                     - self.inst.index[0])

                            if delta < self.orbit_period:
                                # This orbits end occurs on the next day,
                                # though we grabbed the first orbit, missing
                                # data means the first available orbit in the
                                # datais actually the last for the day.
                                # Resetting to the second to last orbit and t
                                # hen callingnext() will get the last orbit,
                                # accounting for tomorrow's data as well.
                                self._current = self.num - 1
                                self.next()
                else:
                    # There is no data for the next day, continue loading data
                    # until there is some.  The `next` method raises
                    # StopIteration when it reaches the end, leaving this
                    # function
                    while self.inst.empty:
                        self.inst.next()
                    self._get_basic_orbit(1)

                del temp_orbit_data
                self._report_current_orbit()

            elif self._current == 0:
                # No current orbit set, grab the first one using the load
                # command to specify the first orbit, which automatically
                # loads prev day if needed to form a complete orbit
                self.load(1)

            elif self._current < (self.num - 1):
                # Since we aren't close to the last orbit, just pull the next
                # orbit
                self._get_basic_orbit(self._current + 1)
                self._report_current_orbit()
            else:
                raise RuntimeError(' '.join(('This is a serious bug. Talk to ',
                                             'someone about this fundamental ',
                                             'failure or open an issue at',
                                             'www.github.com/pysat/pysat')))

        else:
            # There is no data
            while self.inst.empty:
                # Keep going until data is found or next raises stopIteration
                # at the end of the data set, and no more data is available
                self.inst.next()

            # We've found data, grab the next orbit
            self.next()

        return

    def prev(self):
        """Load the previous orbit into associated Instrument.data object

        Note
        ----
        Forms complete orbits across day boundaries. If no data loaded
        then the last orbit of data from the last day is loaded.

        """

        # First, check if data exists
        if not self.inst.empty:
            # Set up orbit metadata
            self._calc_orbits()

            if (self._current > 2) and (self._current <= self.num):
                # If not close to the first orbit, just pull the previous orbit
                #
                # Load orbit and put it into self.inst.data
                self._get_basic_orbit(self._current - 1)
                self._report_current_orbit()
            elif self._current == 2:
                # If current orbit near the first, must be careful
                #
                # First, load prev orbit data
                self._get_basic_orbit(self._current - 1)

                load_prev = True
                if self.inst._iter_type == 'date':
                    delta = self.inst.index[-1] - self.inst.date
                    if delta >= self.orbit_period:
                        # Don't need to load the prev day because this orbit
                        # ends more than a orbital period from start of today's
                        # date
                        load_prev = False

                if load_prev:
                    # Need to save this current orbit and load the prev day
                    temp_orbit_data = self.inst[self.inst.date:]

                    # Load previous day, which clears orbit breaks info
                    try:
                        self.inst.prev()
                        # Combine this next day orbit with previous last orbit
                        if not self.inst.empty:
                            self.inst.concat_data(temp_orbit_data,
                                                  prepend=False)

                            # Select first orbit of combined data
                            self._get_basic_orbit(-1)
                        else:
                            self.inst.next()
                            self._get_basic_orbit(1)
                    except StopIteration:
                        # If loading the first orbit, of first day of data,
                        # you'll end up here as the attempt to make a full
                        # orbit will move the date backwards, and StopIteration
                        # is made. everything is already ok, just move along
                        pass

                    del temp_orbit_data

                self._report_current_orbit()
            elif self._current == 0:
                self.load(-1)
                return
            elif self._current < 2:
                # First, load prev orbit data
                self._get_basic_orbit(1)

                # Need to save this current orbit and load the prev day
                temp_orbit_data = self.inst[self.inst.date:]

                # Load previous day, which clears orbit breaks info
                self.inst.prev()

                # Combine this next day orbit with previous last orbit
                if not self.inst.empty:
                    load_prev = True
                    if self.inst._iter_type == 'date':
                        delta = (self.inst.date - self.inst.index[-1]
                                 + pds.Timedelta('1 day'))
                        if delta >= self.orbit_period:
                            # Don't need to load the prev day because this
                            # orbit ends more than a orbital period from start
                            # of today's date
                            load_prev = False

                    if load_prev:
                        self.inst.concat_data(temp_orbit_data, prepend=False)

                        # Select second to last orbit of combined data
                        self._get_basic_orbit(-2)
                    else:
                        # Padding from the previous is needed
                        self._get_basic_orbit(-1)
                        if self.inst._iter_type == 'date':
                            delta = (self.inst.date - self.inst.index[-1]
                                     + pds.Timedelta('1 day'))
                            if delta < self.orbit_period:
                                self._current = self.num
                                self.prev()
                else:
                    while self.inst.empty:
                        self.inst.prev()
                    self._get_basic_orbit(-1)

                del temp_orbit_data
                self._report_current_orbit()
            else:
                raise RuntimeError(' '.join(('You ended up where nobody should',
                                             'ever be. Talk to someone about',
                                             'this fundamental failure or open',
                                             'an issue at',
                                             'www.github.com/pysat/pysat')))
        else:
            # No data found
            while self.inst.empty:
                # Cycle to more data or raise stopIteration at end of data set
                self.inst.prev()
            self.prev()

        return
